from ...tree_parser.parser_driver import ParserDriver
from ...utils import postprocessor
from ..SDFG.SDFG import DfgRda


class DFGDriver:
    def __init__(
        self,
        src_language="java",
        src_code="",
        output_file="DFG_output.json",
        properties={},
    ):
        self.src_language = src_language

        self.parser = ParserDriver(src_language, src_code).parser
        self.root_node = self.parser.root_node
        self.src_code = self.parser.src_code
        self.properties = properties.get("DFG", {})
        if self.properties.get("statements", False):
            result = DfgRda(
                src_language=src_language,
                src_code=src_code,
                properties=properties,
            )
            self.rda_table = result.rda_table
            self.rda_result = result.rda_result
            self.graph = result.graph
        if output_file:
            self.json = postprocessor.write_networkx_to_json(self.graph, output_file)
            postprocessor.write_to_dot(
                self.graph,
                # os.path.join(grandparent_folder,"output_graphs/AST_output.dot"),
                output_file.split(".")[0] + ".dot",
                output_png=True,
            )


statement_types = {
    "node_list_type": [
        "declaration",
        "expression_statement",
        "labeled_statement",
        "if_statement",
        "while_statement",
        "for_statement",
        "enhanced_for_statement",
        "assert_statement",
        "do_statement",
        "break_statement",
        "continue_statement",
        "return_statement",
        "yield_statement",
        "switch_expression",
        "synchronized_statement",
        "local_variable_declaration",
        "throw_statement",
        "try_statement",
        "try_with_resources_statement",
        "method_declaration",
        "constructor_declaration",
        "switch_block_statement_group",
        "switch_rule",
        "throw_statement",
        "explicit_constructor_invocation",
        "class_declaration",
        "field_declaration",
        "import_declaration",
        "lambda_expression",
        "interface_declaration"
    ],
    "non_control_statement": [
        "declaration",
        "expression_statement",
        "local_variable_declaration",
        "explicit_constructor_invocation",
        "assert_statement",
        "field_declaration",
        "import_declaration"
    ],  # ON extending support fo files, explicit constructor invocation needs to be handled
    "control_statement": [
        "if_statement",
        "while_statement",
        "for_statement",
        "enhanced_for_statement",
        "do_statement",
        "break_statement",
        "continue_statement",
        "return_statement",
        "yield_statement",
        "switch_expression",
        "synchronized_statement",
        "try_statement",
        "try_with_resources_statement",
        "yield_statement",
    ],
    "loop_control_statement": [
        "while_statement",
        "for_statement",
        "enhanced_for_statement",
    ],
    "not_implemented": [],
    "inner_node_type": [
        "declaration",
        "expression_statement",
        "local_variable_declaration",
    ],
    "outer_node_type": ["for_statement"],
    "statement_holders": [
        "block",
        "switch_block_statement_group",
        "switch_block",
        "constructor_body",
        "class_body",
        "module_body",
        "program"
    ],
    "definition_types": ["method_declaration", "constructor_declaration", "class_declaration", "field_declaration", "interface_declaration"]
}

method_return_types = ['integral_type', 'void_type', 'type_identifier', 'generic_type', 'scoped_type_identifier', 'floating_point_type', 'boolean_type', 'array_type']

def get_child_of_type(node, type_list):
    out = list(filter(lambda x : x.type in type_list, node.children))
    if len(out) > 0:
        return out[0]
    else: 
        return None
    
def return_switch_child(node):
    # Make it breadthfirst search, and return if you hit a node_list_type
    bfs_queue = []
    for child in node.children:
        bfs_queue.append(child)
    while bfs_queue != []:
        top = bfs_queue.pop(0)
        if top.type == "switch_expression":
            return top
        if top.type in statement_types["node_list_type"]:
            return None
        for child in top.children:
            bfs_queue.append(child)
    return None

def return_switch_parent(node, non_control_statement):
    """Searches for a switch expression while going up the tree and returns it"""
    
    while node.parent is not None and (node.parent.type != "class_body" and node.parent.type != "method_body"):
        if node.parent.type == "block" and node.type == "switch_expression":
            return node
        if node.parent.type in non_control_statement:
            return node.parent
        node = node.parent
    return None

def return_switch_parent_statement(node, non_control_statement):
    """Searches for a switch expression while going up the tree and returns it"""
    
    while node.parent is not None and (node.parent.type != "class_body" and node.parent.type != "method_body"):
        if node.parent.type in non_control_statement:
            return node.parent
        node = node.parent
    return None

def has_inner_definition(node):
    """Checks if a node has a definition inside it"""
    if node.type in statement_types["definition_types"]:
        return True
    for child in node.children:
        if has_inner_definition(child):
            return True
    return False
def find_method_declaration(node):
    """Searches for a method declaration while going up the tree and returns it"""
    while node.parent is not None:
        if node.parent.type == "method_declaration":
            return node.parent
        node = node.parent
    return None
# def get_last_node(node):
#     """Returns the last statement in the method block"""

def get_signature(node):
    signature = []
    formal_parameters = node.child_by_field_name('parameters')
    formal_parameters = list(filter(lambda x: x.type == 'formal_parameter', formal_parameters.children))
    for formal_parameter in formal_parameters:
        for child in formal_parameter.children:
            if child.type != "identifier":
                signature.append(child.text.decode('utf-8'))
    return tuple(signature)

def get_anonymous_class(node):
    """Checks if a node contains an anonymous class and returns it breadthfirst"""
    bfs_queue = []
    bfs_queue.append(node)
    while bfs_queue != []:
        top = bfs_queue.pop(0)
        if top.type == "class_body" and top.parent.type == "object_creation_expression":
            return top
        for child in top.children:
            bfs_queue.append(child)
    return None

def check_anonymous_class(node):
    """Checks if a node contains a lambda expression"""
    if get_anonymous_class(node) is None:
        return False
    else:
        initial_node = node
        class_node = get_anonymous_class(node)
        parent_node = class_node.parent
        while parent_node is not None:
            if parent_node.type in statement_types["node_list_type"]:            
                if parent_node == initial_node:
                    return True
                else:
                    return False
            parent_node = parent_node.parent

def get_lambda_body_depth(node):
    """Returns the body of a lambda expression"""
    for child in node.children:
        if get_lambda_body(child) is not None:
            return get_lambda_body(child)
        elif child.type == "lambda_expression":
            return child
    return None

def get_lambda_body(node):
    """Returns the body of a lambda expression, breadthfirst"""
    bfs_queue = []

    bfs_queue.append(node)
    while bfs_queue != []:
        top = bfs_queue.pop(0)
        if top.type == "lambda_expression":
            return top
        for child in top.children:
            bfs_queue.append(child)
    return None

def get_all_lambda_body(node):
    """Returns the body of a lambda expression, breadthfirst"""
    bfs_queue = []
    output = []
    bfs_queue.append(node)
    while bfs_queue != []:
        top = bfs_queue.pop(0)
        if top.type == "lambda_expression":
            output.append(top)
        for child in top.children:
            if child.type == "lambda_expression" or child.type not in statement_types["node_list_type"]:
                bfs_queue.append(child)

    return output

def check_lambda(node):
    """Checks if a node contains a lambda expression"""
    if get_lambda_body(node) is None:
        return False
    else:
        initial_node = node
        lambda_node = get_lambda_body(node)
        parent_node = lambda_node.parent
        while parent_node is not None:
            if parent_node.type in statement_types["node_list_type"]:
                
                if parent_node == initial_node:
                    return True
                else:
                    return False
            parent_node = parent_node.parent

def abstract_method(node):
    method_body = get_child_of_type(node, ["block"])
    while node is not None:
        if node.type == "class_body" and node.parent.type == "class_declaration":
            node = node.parent

            modifiers = get_child_of_type(node, ["modifiers"])
            if modifiers is not None and "abstract" in modifiers.text.decode("utf-8") and method_body is None:
                return True
        node = node.parent
    return False

def get_class_name(node, index):
    "Returns the class name when a method declaration or constructor declaration is passed to it"
    type_identifiers = ["type_identifier", "generic_type", "scoped_type_identifier"]
    while node is not None:
        if node.type == "class_body" and node.parent.type == "class_declaration":
            node = node.parent
        
            class_index = index[(node.start_point, node.end_point, node.type)]
            class_name = [(list(filter(lambda child: child.type == "identifier", node.children))[0]).text.decode("UTF-8")]
            
            interface_node = get_child_of_type(node, ["super_interfaces"])
            if interface_node is not None:
                class_name.append(get_child_of_type(interface_node, ["type_list"]).text.decode("UTF-8"))

            superclass_node = get_child_of_type(node, ["superclass"])
            if superclass_node is not None:
                class_name.append(get_child_of_type(superclass_node, ["type_identifier"]).text.decode("UTF-8"))

            return class_index, class_name
        
        elif node.type == "class_body" and node.parent.type == "object_creation_expression":
            node = node.parent
            class_index = index[(node.start_point, node.end_point, node.type)]
            class_name = [list(filter(lambda child: child.type in type_identifiers, node.children))[0].text.decode("UTF-8")]
            return class_index, class_name
        elif node.type == "interface_body" and node.parent.type == "interface_declaration":
            return None
            node = node.parent
            class_index = index[(node.start_point, node.end_point, node.type)]
            class_name = list(filter(lambda child: child.type == "identifier", node.children))[0]
            return class_index, class_name.text.decode("UTF-8")
        node = node.parent

def get_nodes(root_node=None, node_list={}, graph_node_list=[], index={}, records={}):
    """
    Returns statement level nodes recursively from the concrete syntax tree passed to it. Uses records to maintain required supplementary information.
    noe_list maintains an intermediate representation and graph_node_list returns the final list.
    """

    if (
        root_node.type == "parenthesized_expression"
        and root_node.parent is not None
        and root_node.parent.type == "do_statement"
    ):
        label = "while" + root_node.text.decode("UTF-8")
        type_label = "while"
        node_list[(root_node.start_point, root_node.end_point, root_node.type)] = root_node
        graph_node_list.append((index[(root_node.start_point, root_node.end_point, root_node.type)],root_node.start_point[0],label,type_label))

    elif root_node.type == "catch_clause":
        node_list[(root_node.start_point, root_node.end_point, root_node.type)] = root_node
        catch_parameter = list(filter(lambda child: child.type == "catch_formal_parameter", root_node.children))
        label = "catch (" + catch_parameter[0].text.decode("UTF-8") + ")"
        type_label = "catch"
        graph_node_list.append((index[(root_node.start_point, root_node.end_point, root_node.type)],root_node.start_point[0],label,type_label))

    elif root_node.type == "finally_clause":
        node_list[(root_node.start_point, root_node.end_point, root_node.type)] = root_node
        label = "finally"
        type_label = "finally"
        graph_node_list.append((index[(root_node.start_point, root_node.end_point, root_node.type)],root_node.start_point[0],label,type_label))

    elif root_node.type in statement_types["node_list_type"]:
        if (
            root_node.type in statement_types["inner_node_type"]
            and root_node.parent is not None
            and root_node.parent.type in statement_types["outer_node_type"]
            and root_node.parent.child_by_field_name("body") != root_node
        ):
            pass
            # If it has a parent and the parent is a for loop type and it is an initialization or update statement, omit it
        elif (
            root_node.type in statement_types["inner_node_type"]
            and return_switch_child(root_node) is not None
        ):
            # There is a switch expression in the subtree starting from this statement_node
            switch_child = return_switch_child(root_node)
            child_index = index[(switch_child.start_point, switch_child.end_point, switch_child.type)]
            current_index = index[(root_node.start_point, root_node.end_point, root_node.type)]
            records["switch_child_map"][current_index] = child_index
        else:
            node_list[(root_node.start_point, root_node.end_point, root_node.type)] = root_node
            # Set default label values for the node and then modify based on node type if required in the following if-else ladder
            label = root_node.text.decode("UTF-8")
            type_label = "expression_statement"
            if check_anonymous_class(root_node) and root_node.type not in statement_types["definition_types"]:
                try:
                    label = label.split("{")[0] + label.split("}")[-1]
                except:
                    print("INCORRECT anonymous class label", label)
            # elif check_lambda(root_node) and root_node.type not in statement_types["definition_types"]:
            elif check_lambda(root_node) and root_node.type not in statement_types["definition_types"]:
                raw_label = root_node.text.decode("utf-8")

                label = ""
                for lambda_expression in get_all_lambda_body(root_node):
                    split_label = raw_label.split(lambda_expression.text.decode("utf-8"),2)
                    raw_label = split_label[0]
                    if len(split_label) >1:
                        label = split_label[1] + label
                    lambda_node = (lambda_expression.start_point, lambda_expression.end_point, lambda_expression.type)
                    records["lambda_map"][lambda_node] = root_node
                label = raw_label + label

            elif root_node.type == "lambda_expression":
                try:
                    if "{" in label:
                        label = label.split("{")[0] + label.split("}")[-1]
                    else:
                        label = root_node.text.decode('utf-8')
                except:
                    raise Exception("INCORRECT lambda expression label", label)
                    print("INCORRECT lambda expression label", label)

            elif (root_node.type == "method_declaration" or root_node.type == "constructor_declaration"):
                label = ""
                for child in root_node.children:
                    if child.type != "block" and child.type != "constructor_body":
                        label = label + " " + child.text.decode('utf-8')
                method_name = list(
                    filter(lambda child: child.type == "identifier", root_node.children)
                )
                method_name = method_name[0].text.decode("UTF-8")
                method_index = index[(root_node.start_point, root_node.end_point, root_node.type)]
                type_label = root_node.type
                try:
                    signature = get_signature(root_node)
                    class_index, class_name_list = get_class_name(root_node, index)
                    if method_name == "main":
                        records["main_method"] = method_index
                        records["main_class"] = class_index
                    if abstract_method(root_node) == False:
                        for class_name in class_name_list:
                            if root_node.type == "constructor_declaration":
                                if class_name == method_name:
                                    records["constructor_list"][((class_name,method_name), signature)] = method_index
                                return_type = None
                            else:
                                records["method_list"][((class_name,method_name), signature)] = method_index
                                try:
                                    return_type = list(filter(lambda child: child.type in method_return_types, root_node.children))
                                    return_type = return_type[0].text.decode("UTF-8")
                                except:
                                    raise Exception("No return type found for method, update method_return_types", method_name)
                            
                            records["return_type"][((class_name,method_name), signature)] = return_type

                except:
                    pass
                graph_node_list.append((method_index, root_node.start_point[0],label,type_label))
            elif (
                root_node.type == "class_declaration"
                # or root_node.type == "constructor_declaration"
            ):
                modifiers = list(filter(lambda child: child.type != "class_body", root_node.children))
                class_name = list(filter(lambda child: child.type == "identifier", root_node.children))[0].text.decode("UTF-8")
                label = ""
                for modifier in modifiers:
                    label = label + modifier.text.decode("UTF-8") + " "
                type_label = root_node.type
                class_index = index[(root_node.start_point, root_node.end_point, root_node.type)]
                records["class_list"][class_name] = class_index

                superclass_node = get_child_of_type(root_node, ["superclass"])
                if superclass_node is not None:
                    parent_name = get_child_of_type(superclass_node, ["type_identifier", "generic_type", "scoped_type_identifier"]).text.decode("UTF-8")
                    try:
                        records['extends'][class_name].append(parent_name)
                    except:
                        records['extends'][class_name] = [parent_name]

            
            elif root_node.type == "interface_declaration":
                definition = list(
                    filter(
                        lambda child: child.type != "interface_body",
                        root_node.children,
                    )
                )
                label = ""
                for x in definition:
                    label = label + x.text.decode("UTF-8") + " "
                
                type_label = "interface_declaration"

            elif root_node.type == "if_statement":
                condition = list(
                    filter(
                        lambda child: child.type == "parenthesized_expression",
                        root_node.children,
                    )
                )
                label = "if" + condition[0].text.decode("UTF-8")
                type_label = "if"

            elif root_node.type == "for_statement":
                try:
                    init = root_node.child_by_field_name("init").text.decode("UTF-8")
                    if init[-1] != ";":
                        init = init + ";"
                except:
                    init = ""
                try:
                    condition = root_node.child_by_field_name("condition").text.decode(
                        "UTF-8"
                    )
                except:
                    condition = ""
                try:
                    update = root_node.child_by_field_name("update").text.decode(
                        "UTF-8"
                    )
                except:
                    update = ""
                label = "for(" + init + condition + ";" + update + ")"
                type_label = "for"

            elif root_node.type == "enhanced_for_statement":
                try:
                    modifiers = str(
                        list(
                            filter(
                                lambda child: child.type == "modifiers",
                                root_node.children,
                            )
                        )
                    )
                    modifier = modifiers[0].text.decode("UTF-8")
                except:
                    modifier = ""
                try:
                    types = root_node.child_by_field_name("type").text.decode("UTF-8")
                except:
                    types = ""

                try:
                    variables = list(
                        filter(
                            lambda child: child.type == "identifier", root_node.children
                        )
                    )
                    variable = variables[0].text.decode("UTF-8")
                except:
                    variable = ""
                try:
                    value = root_node.child_by_field_name("value").text.decode("UTF-8")
                except:
                    value = ""
                label = (
                    "for(" + modifier + " " + types + " " + variable + ":" + value + ")"
                )
                type_label = "for"

            elif root_node.type == "while_statement":
                condition = list(
                    filter(
                        lambda child: child.type == "parenthesized_expression",
                        root_node.children,
                    )
                )
                label = "while" + condition[0].text.decode("UTF-8")
                type_label = "while"

            elif root_node.type == "do_statement":
                label = "do"
                type_label = "do"

            elif root_node.type == "switch_expression":
                parent_statement = return_switch_parent(root_node, statement_types["non_control_statement"])
                if parent_statement is not None:
                    label = parent_statement.text.decode("UTF-8").split("{")[0]
                else:
                    condition = list(filter(lambda child: child.type == "parenthesized_expression",root_node.children))
                    label = "switch" + condition[0].text.decode("UTF-8")
                type_label = "switch"

            elif (
                root_node.type == "switch_block_statement_group"
                or root_node.type == "switch_rule"
            ):
                case_label = list(
                    filter(
                        lambda child: child.type == "switch_label", root_node.children
                    )
                )
                label = case_label[0].text.decode("UTF-8") + ":"
                type_label = "case"

            elif (
                root_node.type == "try_statement"
                or root_node.type == "try_with_resources_statement"
            ):
                label = "try"
                type_label = "try"

            elif root_node.type == "synchronized_statement":
                condition = list(
                    filter(
                        lambda child: child.type == "parenthesized_expression",
                        root_node.children,
                    )
                )
                label = "synchronized " + condition[0].text.decode("UTF-8")
                type_label = "synchronized"
            elif root_node.type == "labeled_statement":
                name = list(
                    filter(lambda child: child.type == "identifier", root_node.children)
                )
                label = name[0].text.decode("UTF-8") + ":"
                records["label_statement_map"][label] = (
                    root_node.start_point,
                    root_node.end_point,
                    root_node.type,
                )
                type_label = "label"
            elif root_node.type == "return_statement":
                if has_inner_definition(root_node):
                    label = "return"
                else:
                    label = root_node.text.decode("UTF-8")
                type_label = "return"
            
            if (
                root_node.type != "method_declaration"
                and root_node.type != "constructor_declaration"
            ):
                graph_node_list.append(
                    (
                        index[
                            (root_node.start_point, root_node.end_point, root_node.type)
                        ],
                        root_node.start_point[0],
                        label,
                        type_label,
                    )
                )

    for child in root_node.children:
        root_node, node_list, graph_node_list, records = get_nodes(
            root_node=child,
            node_list=node_list,
            graph_node_list=graph_node_list,
            index=index,
            records=records,
        )

    return root_node, node_list, graph_node_list, records



from .AST import ASTGraph
from ...tree_parser.parser_driver import ParserDriver
from ...utils import postprocessor


class ASTDriver:
    def __init__(
        self,
        src_language="java",
        src_code="",
        output_file="AST_output.json",
        properties={},
    ):
        self.src_language = src_language
        self.src_code = src_code
        self.properties = properties
        self.parser = ParserDriver(src_language, src_code).parser
        self.root_node = self.parser.root_node
        self.AST = ASTGraph(
            self.src_language,
            self.src_code,
            self.properties,
            self.root_node,
            self.parser,
        )
        self.graph = self.AST.graph
        if output_file:
            self.json = postprocessor.write_networkx_to_json(self.graph, output_file)
            postprocessor.write_to_dot(
                self.graph, output_file.split(".")[0] + ".dot", output_png=True
            )



"""CLI for depend."""
from __future__ import annotations

import sys
from pathlib import Path
from typing import Optional

import typer
from loguru import logger

from .codeviews.combined_graph.combined_driver import CombinedDriver

app = typer.Typer()


@app.callback(invoke_without_command=True)
def main(
        lang: str = typer.Option(..., help="java, cs"),
        code: Optional[str] = typer.Option(None, help="""
    public class Max2 {
        public static void main(String[] args) {
            int x= 3;
            x = x + 3;
            int y = 4;
            y += 1;
        }
    }
    """),
        code_file: Optional[Path] = typer.Option(None, help="./test_file,java"),
        graphs: str = typer.Option("ast,dfg", help="ast, cfg, dfg"),
        output: str = typer.Option("dot", help="all/json/dot (dot generates png as well)"),
        blacklisted: str = typer.Option("", help="Nodes to be removed from the AST"),
        collapsed: bool = typer.Option(False, help="Collapses all occurrences of a variable into one node"),
        statements: bool = typer.Option(True, help="Converts DFG output to statement level and uses RDA"),
        last_def: bool = typer.Option(False, help="Adds last definition information to the DFG"),
        last_use: bool = typer.Option(False, help="Adds last use information to the DFG"),
        throw_parse_error: bool = typer.Option(False, help="Throws an error if the code cannot be parsed"),
        debug: bool = typer.Option(False, help="Enables debug logs"),
):
    """
    Comex

    Generates, customizes and combines multiple source code representations (AST, CFG, DFG)


    """
    if debug:
        level = "DEBUG"
    else:
        level = "WARNING"

    config = {
        "handlers": [{"sink": sys.stderr, "level": level}],
    }
    logger.configure(**config)

    codeviews = {
        "AST": {
            "exists": False,
            "collapsed": collapsed,
            "minimized": bool(blacklisted),
            "blacklisted": blacklisted.split(",")
        },
        "DFG": {
            "exists": False,
            "collapsed": collapsed,
            "minimized": False,
            "statements": statements,
            "last_def": last_def,
            "last_use": last_use
        },
        "CFG": {
            "exists": False,
        }
    }

    if "ast" in graphs.lower():
        codeviews["AST"] = {
            "exists": True,
            "collapsed": collapsed,
            "minimized": bool(blacklisted),
            "blacklisted": blacklisted.split(",")
        }
    if "dfg" in graphs.lower():
        codeviews["DFG"] = {
            "exists": True,
            "collapsed": collapsed,
            "minimized": False,
            "statements": statements,
            "last_def": last_def,
            "last_use": last_use
        }
    if "cfg" in graphs.lower():
        codeviews["CFG"] = {
            "exists": True,
        }

    try:
        if code_file:
            file_handle = open(code_file, "r")
            src_code = file_handle.read()
            file_handle.close()
            CombinedDriver(src_language=lang, src_code=src_code, output_file="output.json",
                           graph_format=output, codeviews=codeviews)
        else:
            if not code:
                raise Exception("No code provided")
            CombinedDriver(src_language=lang, src_code=code, output_file="output.json", graph_format=output,
                           codeviews=codeviews)
    except (
            Exception
    ) as e:
        try:
            logger.error(e.msg)
        except AttributeError:
            logger.error(e)
        sys.exit(-1)



import traceback

import networkx as nx
from loguru import logger

from .CFG import CFGGraph
from ...utils import cs_nodes


def get_index(node, index):
    return index[(node.start_point, node.end_point, node.type)]


def recursively_get_children_of_types(
        node, st_types, check_list=None, index=None, result=None
):
    if result is None:
        result = []
    if check_list and index:
        result.extend(
            list(
                filter(
                    lambda child: child.type in st_types
                                  and get_index(child, index) in check_list,
                    node.children,
                )
            )
        )
    else:
        result.extend(list(filter(lambda child: child.type in st_types, node.children)))
    for child in node.children:
        result = recursively_get_children_of_types(
            child, st_types, check_list, index, result=result
        )
    return result


class CFGGraph_csharp(CFGGraph):
    def __init__(self, src_language, src_code, properties, root_node, parser):
        super().__init__(src_language, src_code, properties, root_node, parser)

        self.node_list = None
        self.statement_types = cs_nodes.statement_types

        self.CFG_node_list = []
        self.CFG_edge_list = []
        self.records = {
            "basic_blocks": {},
            "method_list": {},
            "constructor_list": {},
            "return_type": {},
            "class_list": {},
            "extends": {},
            "function_calls": {},
            "constructor_calls": {},
            "object_instantiate": {},
            "label_statement_map": {},
            "return_statement_map": {},
            "label_switch_map": {},
            "switch_child_map": {},
            "switch_equivalent_map": {},
        }

        self.index_counter = max(self.index.values())
        self.CFG_node_list, self.CFG_edge_list = self.CFG_cs()
        self.graph = self.to_networkx(self.CFG_node_list, self.CFG_edge_list)

    def get_basic_blocks(self, CFG_node_list, CFG_edge_list):
        G = self.to_networkx(CFG_node_list, CFG_edge_list)
        components = nx.weakly_connected_components(G)
        # NOTE: May need to sort these components according to line number (this one is more foolproof but harder to implement) or the first element in the set (less foolproof, but I think it can be proved, easier to implement).
        # As of now coincidentally the AST node numbering and the list of nodes in networkx are according ot the line numbers
        block_index = 1
        for block in components:
            block_list = sorted(list(block))
            self.records["basic_blocks"][block_index] = block_list
            block_index += 1

    def get_key(self, val, dictionary):
        for key, value in dictionary.items():
            if val in value:
                return key

    def append_block_index(self, CFG_node_list):
        new_list = []
        for node in CFG_node_list:
            block_index = self.get_key(node[0], self.records["basic_blocks"])
            new_list.append((node[0], node[1], node[2], node[3], block_index))
        return new_list

    def add_edge(self, src_node, dest_node, edge_type):
        if src_node == None or dest_node == None:
            logger.error(
                "Node where adding edge is attempted is none {}->{}",
                src_node,
                dest_node,
            )
            logger.warning(traceback.format_stack()[-2])
            raise NotImplementedError
        else:
            self.CFG_edge_list.append((src_node, dest_node, edge_type))

    def get_next_index(self, node_key, node_value):

        next_node = node_value.next_named_sibling
        if next_node == None:
            # Sibling not found

            current_node = node_value
            while current_node.parent is not None:
                if (
                        current_node.parent.type
                        in self.statement_types["loop_control_statement"]
                ):
                    next_node = current_node.parent
                    break
                next_node = current_node.next_named_sibling
                if next_node is not None:
                    for i, child in enumerate(next_node.parent.children):
                        # if child.type == "while" and child.parent.type == "do_statement":
                        #     return None
                        if (
                                child.type == "while"
                                and child.parent.type == "do_statement"
                        ):
                            while_node = child.next_named_sibling
                            next_node = while_node
                            break
                        if (
                                child.type in ["catch_clause", "finally_clause"]
                                and child.parent.type == "try_statement"
                        ):
                            next_node = None
                            break
                        if child == next_node:
                            field_name = next_node.parent.field_name_for_child(i)
                            if field_name is None:
                                break
                            else:
                                next_node = None
                                break
                if next_node is not None:
                    break
                current_node = current_node.parent

        if next_node == None:
            return 2
        if next_node.type == "block":
            for child in next_node.children:
                if child.is_named:
                    next_node = child
                    break

        return self.index[(next_node.start_point, next_node.end_point, next_node.type)]

    def edge_first_line(self, current_node_key, current_node_value):
        # We need to add an edge to the first statement in the next basic block
        node_index = self.index[current_node_key]
        try:
            current_block_index = self.get_key(node_index, self.records["basic_blocks"])
            next_block_index = current_block_index + 1
            first_line_index = self.records["basic_blocks"][next_block_index][0]
            src_node = node_index
            dest_node = first_line_index
            self.add_edge(
                src_node, dest_node, "first_next_line"
            )  # We could maybe differentiate this
        except:
            # Most probably the block is empty
            # add a direct edge to the next statement
            # print("HIT AN EMPTY BLOCK")
            next_index = self.get_next_index(current_node_key, current_node_value)
            self.add_edge(node_index, next_index, "next_line")

    def edge_to_body(self, current_node_key, current_node_value, body_type, edge_type):
        # We need to add an edge to the first statement in the body block
        src_node = self.index[current_node_key]
        # [current_node_value.field_name_for_child(i) for i,child in enumerate(current_node_value.children)]
        body_node = current_node_value.child_by_field_name(body_type)
        if body_node is None:
            body_nodes = [
                child for child in current_node_value.children if child.type == "block"
            ]
            if body_nodes:
                body_node = body_nodes[0]
            else:
                body_node = current_node_value.named_children[-1]
        flag = False
        while body_node.type == "block":
            for child in body_node.children:
                if child.is_named:
                    flag = True
                    body_node = child
                    break
            if flag is False:
                return
        if (
                body_node.is_named
                and body_node.type in self.statement_types["node_list_type"]
        ):
            dest_node = self.index[
                (body_node.start_point, body_node.end_point, body_node.type)
            ]
            self.add_edge(src_node, dest_node, edge_type)

    def get_block_last_line(self, current_node_value, block_type):
        # Find the last line in the body block
        block_node = current_node_value.child_by_field_name(block_type)
        if block_node is None:
            block_node = [
                node
                for node in current_node_value.named_children
                if node.type == block_type
            ]
            if len(block_node) == 0:
                for child in reversed(current_node_value.children):
                    if child.is_named:
                        block_node = child
                        break
            else:
                block_node = block_node[0]

        if block_node.is_named is False:
            return (
                self.index[
                    (
                        current_node_value.start_point,
                        current_node_value.end_point,
                        current_node_value.type,
                    )
                ],
                current_node_value.type,
            )

        while block_node.type == "block":
            named_children = list(
                filter(
                    lambda child: child.is_named == True, reversed(block_node.children)
                )
            )
            if len(named_children) == 0:
                # It means there is an empty block - thats why no named nodes inside
                return (
                    self.index[
                        (
                            current_node_value.start_point,
                            current_node_value.end_point,
                            current_node_value.type,
                        )
                    ],
                    current_node_value.type,
                )
            block_node = named_children[0]
            if block_node.type in self.statement_types["node_list_type"]:
                return (
                    self.index[
                        (block_node.start_point, block_node.end_point, block_node.type)
                    ],
                    block_node.type,
                )
        return (
            self.index[(block_node.start_point, block_node.end_point, block_node.type)],
            block_node.type,
        )

    def function_list(self, current_node):
        if current_node.type == "method_invocation":
            # maintain a list of all method invocations
            method_name = current_node.child_by_field_name("name").text.decode("UTF-8")

            parent_node = None
            pointer_node = current_node
            while pointer_node is not None:
                if (
                        pointer_node.parent is not None
                        and pointer_node.parent.type
                        in self.statement_types["node_list_type"]
                ):
                    parent_node = pointer_node.parent
                    break
                pointer_node = pointer_node.parent

            # Removing this if condition will treat all print sttements as function calls as well
            if method_name != "println" and method_name != "print":
                # index : (AST_id, method_name) (AST_id is of the parent node)
                if method_name in self.records["function_calls"].keys():
                    self.records["function_calls"][method_name].append(
                        (
                            self.index_counter,
                            self.index[
                                (
                                    parent_node.start_point,
                                    parent_node.end_point,
                                    parent_node.type,
                                )
                            ],
                        )
                    )
                else:
                    self.index_counter += 1
                    self.records["function_calls"][method_name] = [
                        (
                            self.index_counter,
                            self.index[
                                (
                                    parent_node.start_point,
                                    parent_node.end_point,
                                    parent_node.type,
                                )
                            ],
                        )
                    ]
                    # Patent node of function call AST id maps to AST id or index of dummy external funciton call node
                    # self.records['function_calls'][index] = (self.index_counter, method_name)
                    if method_name not in self.records["method_list"].keys():
                        self.CFG_node_list.append(
                            (
                                self.index_counter,
                                0,
                                "function_call: " + method_name,
                                "external_function",
                            )
                        )

        for child in current_node.children:
            if child.is_named:
                self.function_list(child)

    def add_dummy_nodes(self):
        self.CFG_node_list.append((1, 0, "start_node", "start"))
        self.CFG_node_list.append((2, 0, "exit_node", "exit"))

    def add_dummy_edges(self):
        for node_name, node_index in self.records["function_calls"].items():
            for node in node_index:
                if node_name not in self.records["method_list"].keys():
                    self.add_edge(node[1], node[0], "function_call")
                    self.add_edge(node[0], node[1], "function_return")
                else:
                    self.add_edge(
                        node[1],
                        self.records["method_list"][node_name],
                        "recursive_method_call",
                    )

    def read_index(self, index):
        return list(self.index.keys())[list(self.index.values()).index(index)]

    def CFG_cs(self):
        warning_counter = 0
        node_list = {}
        # node_list is a dictionary that maps from (node.start_point, node.end_point, node.type) to the node object
        # of tree-sitter
        _, self.node_list, self.CFG_node_list, self.records = cs_nodes.get_nodes(
            root_node=self.root_node,
            node_list=node_list,
            graph_node_list=self.CFG_node_list,
            index=self.index,
            records=self.records,
        )
        # for eq_case, cases in self.records["switch_equivalent_map"].items():
        #     equivalent = [eq_case] + cases
        #     for k,v in self.records["label_statement_map"].items():
        #         if v in equivalent:
        #             for case in equivalent:
        #                 if case not in self.records["label_statement_map"]:
        #                     self.records["label_statement_map"]

        # Initial for loop required for basic block creation and simple control flow within a block ----------------------------
        for node_key, node_value in node_list.items():
            current_node_type = node_key[2]
            if current_node_type in self.statement_types["non_control_statement"]:
                # if current_node_type not in self.statement_types['terminal_inner']:
                if cs_nodes.return_switch_child(node_value) is None:
                    try:
                        src_node = self.index[node_key]
                        if node_value.type == "labeled_statement":
                            if "statement" in node_value.named_children[-1].type:
                                src_node = get_index(
                                    node_value.named_children[-1], self.index
                                )
                        next_node = node_value.next_named_sibling
                        if next_node and next_node.type == "block":
                            for child in next_node.children:
                                if child.type in self.statement_types["node_list_type"]:
                                    self.add_edge(
                                        src_node,
                                        self.index[
                                            (
                                                child.start_point,
                                                child.end_point,
                                                child.type,
                                            )
                                        ],
                                        "into_block",
                                    )
                                    break
                        else:
                            dest_node = self.index[
                                (
                                    next_node.start_point,
                                    next_node.end_point,
                                    next_node.type,
                                )
                            ]
                            if dest_node in self.records["switch_child_map"].keys():
                                dest_node = self.records["switch_child_map"][dest_node]
                            for i, child in enumerate(next_node.parent.children):
                                if child == next_node:
                                    field_name = next_node.parent.field_name_for_child(
                                        i
                                    )
                                    if field_name is None:
                                        self.add_edge(src_node, dest_node, "next_line")
                                    else:
                                        # TODO Triage
                                        next_node = None
                    except:
                        pass

            elif current_node_type in self.statement_types["not_implemented"]:
                print("WARNING: Not implemented ", current_node_type)
                warning_counter += 1

        self.get_basic_blocks(self.CFG_node_list, self.CFG_edge_list)
        self.CFG_node_list = self.append_block_index(self.CFG_node_list)

        self.function_list(self.root_node)

        self.add_dummy_nodes()
        self.add_dummy_edges()
        # ------------------------------------------------------------------------------
        # At this point, the self.CFG_node_list has basic block index appended to it
        # ------------------------------------------------------------------------------
        first_guy_hit = False
        for node_key, node_value in node_list.items():
            current_node_type = node_key[2]
            current_index = self.index[node_key]
            if current_node_type in [
                "method_declaration",
                "interface_declaration",
                "constructor_declaration",
                "property_declaration",
            ]:
                # We need to add an edge to the first statement in the next basic block
                if not first_guy_hit:
                    first_guy_hit = True
                    self.add_edge(1, current_index, "next")
                self.edge_first_line(node_key, node_value)
                last_line_index, line_type = self.get_block_last_line(
                    node_value, "body"
                )
                if line_type in self.statement_types["non_control_statement"]:
                    if last_line_index in self.records["switch_child_map"].keys():
                        last_line_index = self.records["switch_child_map"][
                            last_line_index
                        ]
                    if line_type == "labeled_statement":
                        statement_node = node_list[self.read_index(last_line_index)]
                        if "statement" in statement_node.named_children[-1].type:
                            last_line_index = get_index(
                                statement_node.named_children[-1], self.index
                            )
                    self.add_edge(last_line_index, 2, "exit_next")

            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "lock_statement":
                self.edge_first_line(node_key, node_value)
                last_line_index, line_type = self.get_block_last_line(
                    node_value, "body"
                )
                next_dest_index = self.get_next_index(node_key, node_value)
                if line_type in self.statement_types["non_control_statement"]:
                    self.add_edge(last_line_index, next_dest_index, "lock_released")
                #     self.add_edge(last_line_index, 2, 'exit_next')

            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "labeled_statement":
                self.edge_first_line(node_key, node_value)

            elif current_node_type in [
                "checked_statement",
                "fixed_statement",
                "unsafe_statement",
                "using_statement",
                "local_function_statement",
            ]:
                self.edge_first_line(node_key, node_value)
                last_line_index, line_type = self.get_block_last_line(
                    node_value, "body"
                )
                if line_type in self.statement_types["non_control_statement"]:
                    self.add_edge(
                        last_line_index,
                        self.index[
                            (
                                node_value.start_point,
                                node_value.end_point,
                                node_value.type,
                            )
                        ],
                        "return_control",
                    )
                # next_dest_index = self.get_next_index(node_key, node_value)
                # self.add_edge(self.index[node_key], next_dest_index, 'next_line')

            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "if_statement":
                # Find the if block body and the else block body if exists (first statement inside them, add an edge)
                # Find the line just after the entire if_statement
                next_dest_index = self.get_next_index(node_key, node_value)
                # consequence
                self.edge_to_body(node_key, node_value, "consequence", "pos_next")
                # Find the last line in the consequence block and add an edge to the next statement
                last_line_index, line_type = self.get_block_last_line(
                    node_value, "consequence"
                )
                # Also add an edge from the last guy to the next statement after the if
                # print(last_line_index, line_type)
                if line_type in self.statement_types["non_control_statement"]:
                    self.add_edge(last_line_index, next_dest_index, "next_line")

                # alternative = node_value.child_by_field_name('alternative')
                # print("alternative", alternative)
                if node_value.child_by_field_name("alternative") is not None:

                    # alternative
                    self.edge_to_body(node_key, node_value, "alternative", "neg_next")
                    # Find the last line in the alternative block
                    last_line_index, line_type = self.get_block_last_line(
                        node_value, "alternative"
                    )
                    # print(last_line_index, line_type)
                    if line_type in self.statement_types["non_control_statement"]:
                        self.add_edge(last_line_index, next_dest_index, "next_line")
                else:
                    # When else is not there add a direct edge from if node to the next statement
                    # if node_value.parent.type == "block" and node_value.parent.parent.type == "do_statement":
                    #     pass
                    # TODO FIX LOGIC IMP
                    # elif cs_nodes.return_index_of_first_parent_of_type(node_value, 'switch_section') is not None:
                    #     pass
                    # else:
                    self.add_edge(current_index, next_dest_index, "next_line")

            # ------------------------------------------------------------------------------------------------
            elif current_node_type in self.statement_types["loop_control_statement"]:
                # Get the node immediately after the while statement
                next_dest_index = self.get_next_index(node_key, node_value)

                # Add an edge from this node to the first line in the body
                self.edge_to_body(node_key, node_value, "body", "pos_next")

                # Find the last line in the body block
                last_line_index, line_type = self.get_block_last_line(
                    node_value, "body"
                )

                # Add an edge from this node to the next line after the loop statement
                self.add_edge(current_index, next_dest_index, "neg_next")
                # Add an edge from the last statement in the body to this node
                if line_type in self.statement_types["non_control_statement"]:
                    self.add_edge(last_line_index, current_index, "loop_control")

                # Add a self loop in case of for loops
                if current_node_type != "while_statement":
                    self.add_edge(current_index, current_index, "loop_update")

            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "do_statement":
                # Find the corresponding while statement
                # Add an edge from this node to the first line in the body
                # Find the last statement in the body and an edge frpom last line to the while node
                # Add an edge from the while node to the first line in the block or to the current do node
                # Add an edge from the while node to the next statement after the do_statement

                # Get the node immediately after the while statement
                next_dest_index = self.get_next_index(node_key, node_value)

                # Add an edge from this node to the first line in the body
                self.edge_to_body(node_key, node_value, "body", "pos_next")

                # Find the last line in the body block
                last_line_index, line_type = self.get_block_last_line(
                    node_value, "block"
                )

                # Search the CFG_node_list for parameterized_expression with parent do_statement with AST_id = src_node
                # for k, v in node_list.items():
                #     # print(k,v)
                #     if self.index[
                #         (v.parent.start_point, v.parent.end_point, v.parent.type)] == current_index:
                #         if 'expression' in k[2]:
                #             while_index = self.index[k]
                #             break
                while_index = None
                for child in node_value.children:
                    if child.type == "while":
                        while_node = child.next_named_sibling
                        while_index = self.index[
                            (
                                while_node.start_point,
                                while_node.end_point,
                                while_node.type,
                            )
                        ]
                        break

                # Find the last statement in the body and an edge frpom last line to the while node
                if line_type not in self.statement_types["control_statement"]:
                    self.add_edge(last_line_index, while_index, "next")

                # Add an edge from the while node to the first line in the block or to the current do node
                # self.CFG_edge_list.append((while_node, dest_node, 'loop_control')) # First node of block
                self.add_edge(while_index, current_index, "loop_control")  # do node

                # Add an edge from the while node to the next statement after the do_statement
                self.add_edge(while_index, next_dest_index, "neg_next")

            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "continue_statement":
                # Go up the parent chain until we reach a loop_control statement or do statement
                # add edge from this node to the loop_control statement or do statement

                parent_node = node_value.parent
                loop_node = None
                while parent_node is not None:
                    if (
                            parent_node.type
                            in self.statement_types["loop_control_statement"]
                            or parent_node.type == "do_statement"
                    ):
                        loop_node = parent_node
                        break
                    parent_node = parent_node.parent

                # add edge from this node to the loop_control statement or do statement

                src_node = self.index[node_key]
                dest_node = self.index[
                    (loop_node.start_point, loop_node.end_point, loop_node.type)
                ]
                # label_name = list(filter(lambda child: child.type == 'identifier', node_value.children))
                # if len(label_name) > 0:
                #     target_name = label_name[0].text.decode('UTF-8').strip()
                #     dest_node = self.records['label_statement_map'][target_name]
                self.CFG_edge_list.append((src_node, dest_node, "jump_next"))

            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "break_statement":
                # if it is inside a switch, it is handled here and also refer to switch_expression,
                # if it is inside a loop, handle it here
                parent_node = node_value.parent
                loop_node = None
                while parent_node is not None:
                    if (
                            parent_node.type
                            in self.statement_types["loop_control_statement"]
                            + ["do_statement"]
                            or parent_node.type == "switch_expression"
                            or parent_node.type == "switch_statement"
                    ):
                        loop_node = parent_node
                        break
                    parent_node = parent_node.parent
                if loop_node is not None:
                    next_dest_index = self.get_next_index(
                        (loop_node.start_point, loop_node.end_point, loop_node.type),
                        loop_node,
                    )
                    # label_name = list(filter(lambda child: child.type == 'identifier', node_value.children))
                    # if len(label_name) > 0:
                    #     target_name = label_name[0].text.decode('UTF-8') + ":"
                    #     next_dest_index = self.records['label_statement_map'][target_name]
                    self.add_edge(current_index, next_dest_index, "jump_next")
            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "goto_statement":
                target_name = node_value.text.decode("UTF-8").replace(";", "").replace("goto", "").strip()
                if target_name in self.records["label_statement_map"]:
                    next_dest_index = self.records["label_statement_map"][target_name]
                else:
                    switch_node = cs_nodes.return_index_of_first_parent_of_type(
                        node_value, "switch_statement"
                    )
                    switch_id = self.index[
                        (
                            switch_node.start_point,
                            switch_node.end_point,
                            switch_node.type,
                        )
                    ]
                    next_dest_index = self.records["label_switch_map"][switch_id][
                        target_name
                    ]
                self.add_edge(current_index, next_dest_index, "jump_next")
            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "yield_statement":
                # if it is inside a switch, it is handled here and also refer to switch_expression,
                parent_node = node_value.parent
                loop_node = None
                while parent_node is not None:
                    if (
                            parent_node.type
                            in self.statement_types["loop_control_statement"]
                            or parent_node.type == "do_statement"
                            or parent_node.type == "switch_expression"
                    ):
                        loop_node = parent_node
                        break
                    parent_node = parent_node.parent

                try:
                    next_dest_index = self.index[
                        (loop_node.start_point, loop_node.end_point, loop_node.type)
                    ]
                except:
                    # Handle yield when no loop parent or switch parent
                    next_dest_index = self.get_next_index(
                        (node_value.start_point, node_value.end_point, node_value.type),
                        node_value,
                    )

                self.add_edge(current_index, next_dest_index, "yield_exit")
            # ------------------------------------------------------------------------------------------------
            elif (
                    current_node_type == "switch_expression"
                    or current_node_type == "switch_statement"
            ):
                # First check if the switch expression_statement is part of a non-control statement, and add an edge to the next line
                # if current_node_type == 'switch_expression':
                #     switch_parent = cs_nodes.return_switch_parent(node_value,
                #                                                   self.statement_types['non_control_statement'])
                #     self.add_edge(self.index[(switch_parent.start_point, switch_parent.end_point, switch_parent.type)], current_index, 'next_line')

                # try:
                #     next_node = switch_parent.next_named_sibling
                #     src_node = self.index[node_key]
                #     dest_node = self.index[(next_node.start_point, next_node.end_point, next_node.type)]
                #     if dest_node in self.records['switch_child_map'].keys():
                #         dest_node = self.records['switch_child_map'][dest_node]
                #     self.add_edge(src_node, dest_node, 'next_line')
                #
                # except Exception as e:
                #     print(e)
                #     pass

                # Find all the case blocks associated with this switch node and add an edge to each of them
                # Find the last line in each case block and add an edge to the next case block unless it is a break statement
                # BUt if the block is empty, add an edge to the next case label
                # in case of a break statement, add an edge to the next statement outside the switch
                # in case of default, add an edge to the next statement outside the switch
                # in case of no default, add an edge from last block to the next statement outside the switch
                case_node_list = {}
                # default_exists = False

                # Find the next statement outside the switch
                next_dest_index = self.get_next_index(node_key, node_value)

                if current_index in self.records["label_switch_map"]:
                    if "default" not in self.records["label_switch_map"][current_index]:
                        self.add_edge(
                            current_index, next_dest_index, "switch_no_default"
                        )

                # For each case label block, find the first statement in the block and add an edge to it
                for k, v in node_list.items():
                    # print(k,v)
                    source_index = None
                    if k[2] == "switch_expression_arm" or k[2] == "switch_section":
                        if k[2] == "switch_expression_arm":
                            targ = v.parent
                            if (
                                    self.index[
                                        (targ.start_point, targ.end_point, targ.type)
                                    ]
                                    == current_index
                            ):
                                case_node_index = self.index[k]
                                self.add_edge(
                                    current_index, case_node_index, "switch_case"
                                )
                                case_node_list[k] = v
                        else:
                            targ = v.parent.parent
                            if (
                                    self.index[
                                        (targ.start_point, targ.end_point, targ.type)
                                    ]
                                    == current_index
                            ):
                                source_index = current_index
                                type_edge = "switch_case"
                                sink_index = None
                                for eq_case in self.records["switch_equivalent_map"][
                                    self.index[k]
                                ]:
                                    sink_index = eq_case
                                    self.add_edge(source_index, sink_index, type_edge)
                                    type_edge = "fall through"
                                    source_index = sink_index
                            current_case_index = self.index[k]
                            # case_statements = list(
                            #     filter(lambda child: (child.is_named and child.type != 'case_switch_label'), v.children))
                            # case_statements = v.named_children
                            next_case_node = v.next_named_sibling
                            try:
                                next_case_node_index = self.index[
                                    (
                                        next_case_node.start_point,
                                        next_case_node.end_point,
                                        next_case_node.type,
                                    )
                                ]
                            except:
                                next_case_node_index = None

                            # INVALID IN THE CASE OF C#
                            # if not len(case_statements) > 1:
                            #     # There is no case body, so add an edge from this case to the next case label, if exists
                            #     self.add_edge(current_case_index, next_case_node_index, 'fall through')

                            # else:
                            # The case body exists
                            # Find the first line in each case block and add an edge from case label to it
                            block_node = None
                            for child_node in v.named_children[1:]:
                                # Need to write a loop for unlimited layers of nesting
                                if "block" in child_node.type:
                                    block_node = child_node.named_children[0]
                                    break
                                else:
                                    if "label" not in child_node.type:
                                        block_node = child_node
                                        break
                            first_line_index = self.index[
                                (
                                    block_node.start_point,
                                    block_node.end_point,
                                    block_node.type,
                                )
                            ]
                            if first_line_index in self.records["switch_child_map"]:
                                first_line_index = self.records["switch_child_map"][
                                    first_line_index
                                ]
                            if source_index is not None:
                                self.add_edge(
                                    source_index, first_line_index, "case_next"
                                )

                            # Find the last line in each case block and add an edge to the next case block unless it is a break statement
                            block_node = None
                            for child in reversed(v.children):
                                # Need to write a loop for unlimited layers of nesting
                                if child.is_named:
                                    if child.type == "block":
                                        for child in reversed(child.children):
                                            if child.is_named:
                                                block_node = child
                                                break
                                    else:
                                        block_node = child
                                    break

                            # last_line_index = self.index[
                            #     (block_node.start_point, block_node.end_point, block_node.type)]
                            # if block_node.type in self.statement_types['non_control_statement']:
                            #     # print(block_node.text.decode('UTF-8'))
                            #     self.add_edge(last_line_index, next_case_node_index, 'fall through')

                            # in case of default, add an edge to the next statement outside the switch
                            # in case of no default, add an edge from last block to the next statement outside the switch
                            if next_case_node_index is None:
                                # This is the last block
                                if (
                                        block_node.type
                                        in self.statement_types["non_control_statement"]
                                ):
                                    self.add_edge(
                                        last_line_index, next_dest_index, "switch_out"
                                    )
                            # case_label = list(filter(lambda child : child.type == 'switch_label', v.children))
                            # if case_label == 'default':
                            #     default_exists = True

                for k, v in case_node_list.items():
                    current_case_index = self.index[k]
                    parent = v.parent
                    while parent.parent is not None:
                        parent = parent.parent
                        if parent.type == "local_declaration_statement":
                            break
                    next_dest_index = self.get_next_index(None, parent)
                    self.add_edge(current_case_index, next_dest_index, "switch_out")

                # in case of a break statement, add an edge to the next statement outside the switch
                # -> Handled in break statement

            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "return_statement":
                self.add_edge(current_index, 2, "return_exit")

            # ------------------------------------------------------------------------------------------------
            elif current_node_type == "try_statement":
                # Add edge from try block to first statement inside the block
                self.edge_to_body(node_key, node_value, "body", "next")
                catch_node_list = {}
                finally_node = None
                # Identify all catch blocks - add an edge from catch node to first line inside block
                for k, v in node_list.items():

                    if (
                            k[2] == "catch_clause"
                            and self.index[
                        (v.parent.start_point, v.parent.end_point, v.parent.type)
                    ]
                            == current_index
                    ):
                        catch_node_list[k] = v
                        self.edge_to_body(k, v, "body", "next")
                        # catch_node_index = self.index[k]
                        # self.add_edge(current_index, catch_node_index, 'catch_next')

                    elif (
                            k[2] == "finally_clause"
                            and self.index[
                                (v.parent.start_point, v.parent.end_point, v.parent.type)
                            ]
                            == current_index
                    ):
                        finally_node = v
                        # ? TODO Triage
                        # self.add_edge(current_index, get_index(v, self.index), "always_finally")
                        self.edge_first_line(k, v)  # Not sure if this works

                # From each line inside the try block, an edge going to each catch block
                try_body = node_value.child_by_field_name("body")
                check_list = [node[0] for node in self.CFG_node_list]
                statements = recursively_get_children_of_types(
                    try_body,
                    self.statement_types["node_list_type"],
                    check_list,
                    self.index,
                )

                if len(statements) > 0:
                    for statement in statements:
                        statement_index = self.index[
                            (statement.start_point, statement.end_point, statement.type)
                        ]
                        for catch_node in catch_node_list.keys():
                            if statement.type != "throw_statement":
                                catch_index = self.index[catch_node]
                                # if statement.type != 'return_statement': The Exception can occur on the RHS so the catch_exception edge should be there
                                self.add_edge(
                                    statement_index, catch_index, "catch_exception"
                                )

                # Find the next statement outside the try block
                next_dest_index = self.get_next_index(node_key, node_value)
                exit_next = None

                # finally block is optional
                if finally_node is not None:
                    # From last line of finally to next statement outside the try block
                    last_line_index, line_type = self.get_block_last_line(
                        finally_node, "body"
                    )
                    if line_type in self.statement_types["non_control_statement"]:
                        self.add_edge(last_line_index, next_dest_index, "finally_exit")
                    # For the remaining portion, set finally block as next node if exists
                    exit_next = self.index[
                        (
                            finally_node.start_point,
                            finally_node.end_point,
                            finally_node.type,
                        )
                    ]
                else:
                    exit_next = next_dest_index

                # From last line of try block to finally or to next statement outside the try block
                last_line_index, line_type = self.get_block_last_line(
                    node_value, "body"
                )
                if (
                        line_type in self.statement_types["non_control_statement"]
                        or finally_node is not None
                ):
                    self.add_edge(last_line_index, exit_next, "try_exit")
                # From last line of every catch block to finally or to next statement outside the try block
                for catch_node, catch_value in catch_node_list.items():
                    last_line_index, line_type = self.get_block_last_line(
                        catch_value, "body"
                    )
                    if line_type in self.statement_types["non_control_statement"]:
                        self.add_edge(last_line_index, exit_next, "catch_exit")
                        # Case of empty catch block
                    elif last_line_index == self.index[catch_node]:
                        self.add_edge(last_line_index, exit_next, "catch_exit")
                        # ------------------------------------------------------------------------------------------------
            elif current_node_type == "throw_statement":
                # Control goes to the first catch in the call stack (dynamic call stack) Nothing to do statically
                # Essentially exits the function

                parent = node_value.parent
                try_flag = False
                while parent is not None:
                    if parent.type == "catch_clause" or parent.type == "finally_clause":
                        break
                    if parent.type == "try_statement":
                        if any(
                                [c.type == "catch_clause" for c in parent.named_children]
                        ):
                            try_flag = True
                        break
                    parent = parent.parent
                if try_flag is False:
                    self.add_edge(current_index, 2, "throw_exit")
                else:
                    for child in parent.named_children:
                        if child.type == "catch_clause":
                            self.add_edge(
                                current_index,
                                get_index(child, self.index),
                                "throw_catch",
                            )
            else:
                pass

        if warning_counter > 0:
            print(
                "Total number of warnings from unimplemented statement types: ",
                warning_counter,
            )
        return self.CFG_node_list, self.CFG_edge_list



import shutil
import tempfile

from tree_sitter import Language, Parser
import os
import subprocess

def get_commit_hash(directory):
    try:
        result = subprocess.run(['git', 'rev-parse', 'HEAD'], cwd=directory, capture_output=True, text=True)
        if result.returncode == 0:
            commit_hash = result.stdout.strip()
            return commit_hash
        return None
    except FileNotFoundError:
        return None


class CustomParser:
    """Custom parser for the src_language"""

    def __init__(self, src_language, src_code):
        """Initialize the parser with the language.
        Language options are: python, java, c_sharp, ruby, go, php, javascript"""

        self.src_language = src_language
        self.src_code = src_code
        self.index = {}

        clone_directory = os.path.join(tempfile.gettempdir(), "comex")
        shared_languages = os.path.join(clone_directory, "languages.so")

        grammar_repos = [
            ("https://github.com/tree-sitter/tree-sitter-java", "09d650def6cdf7f479f4b78f595e9ef5b58ce31e"),
            ("https://github.com/tree-sitter/tree-sitter-c-sharp", "3ef3f7f99e16e528e6689eae44dff35150993307")
        ]
        vendor_languages = []

        if not os.path.isfile(shared_languages):
            print("First time running COMEX: Setting up tree-sitter grammars")
            for url, commit in grammar_repos:
                vendor_language = os.path.join(clone_directory, url.rstrip("/").split("/")[-1])
                vendor_languages.append(vendor_language)
                if os.path.exists(vendor_language):
                    shutil.rmtree(vendor_language)
                os.makedirs(vendor_language, exist_ok=True)
                subprocess.check_call(["git", "init"], cwd=vendor_language, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
                subprocess.check_call(["git", "remote", "add", "origin", url], cwd=vendor_language, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
                subprocess.check_call(["git", "fetch", "--depth=1", "origin", commit], cwd=vendor_language, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
                subprocess.check_call(["git", "checkout", commit], cwd=vendor_language, stdout=subprocess.DEVNULL, stderr=subprocess.STDOUT)
        else:
            for url, commit in grammar_repos:
                vendor_language = os.path.join(clone_directory, url.rstrip("/").split("/")[-1])
                vendor_languages.append(vendor_language)

        # build_id = ""
        # for vendor_language in vendor_languages:
        #     commit_hash = get_commit_hash(vendor_language)
        #     if commit_hash:
        #         build_id += commit_hash
        #     else:
        #         build_id += "ERROR"
        # build_id_file = os.path.join(clone_directory, "build_id")
        #
        # # check if the build_id is the same as the one stored in the file
        # # if not, rebuild the shared library
        # if os.path.exists(build_id_file):
        #     with open(build_id_file, "r") as f:
        #         stored_build_id = f.read()
        #     if build_id != stored_build_id:
        #         os.remove(shared_languages)
        # else:
        #     if os.path.exists(shared_languages):
        #         os.remove(shared_languages)

        self.language = Language.build_library(
            # Store the library in the `build` directory
            shared_languages,
            vendor_languages,
        )
        # PYTHON_LANGUAGE = Language("build/my-languages.so", "python")
        JAVA_LANGUAGE = Language(shared_languages, "java")
        C_SHARP_LANGUAGE = Language(shared_languages, "c_sharp")
        # RUBY_LANGUAGE = Language("build/my-languages.so", "ruby")
        # GO_LANGUAGE = Language("build/my-languages.so", "go")
        # PHP_LANGUAGE = Language("build/my-languages.so", "php")
        # JAVASCRIPT_LANGUAGE = Language("build/my-languages.so", "javascript")

        # with open(build_id_file, "w") as f:
        #     f.write(build_id)

        self.language_map = {
            # "python": PYTHON_LANGUAGE,
            "java": JAVA_LANGUAGE,
            "cs": C_SHARP_LANGUAGE,
            # "ruby": RUBY_LANGUAGE,
            # "go": GO_LANGUAGE,
            # "php": PHP_LANGUAGE,
            # "javascript": JAVASCRIPT_LANGUAGE,
        }
        self.root_node, self.tree = self.parse()
        self.all_tokens = []  # list of all tokens in the source code
        self.label = {}  # label for each node in the AST
        self.method_map = []  # list of all methods in the source code
        self.method_calls = []  # list of all method calls in the source code
        self.start_line = {}  # start line number for each toekn (leaf-node) in the AST
        self.declaration = {}  # maps the declaration identifier token's index to the name of the variable/identifier
        self.declaration_map = {}
        self.symbol_table = {
            "scope_stack": [0],
            "scope_map": {},
            "scope_id": 0,
            "data_type": {},
        }  # Stack style symbol table which maps

    def create_AST_id(self, root_node, AST_index, AST_id):
        """Create an id for each node in the AST. This AST id is maintained and used across all code views so that
        it is possible to easily combine graphs"""
        if root_node.is_named:
            current_node_id = AST_id[0]
            AST_id[0] += 1
            AST_index[
                (root_node.start_point, root_node.end_point, root_node.type)
            ] = current_node_id
            for child in root_node.children:
                if child.is_named:
                    self.create_AST_id(child, AST_index, AST_id)
            return

    def parse(self):
        parser = Parser()  # tree-sitter parser
        parser.set_language(self.language_map[self.src_language])
        tree = parser.parse(bytes(self.src_code, "utf8"))
        self.root_node = tree.root_node
        # First few id values are reserved for special nodes such as start and end node
        self.create_AST_id(self.root_node, self.index, [5])
        return self.root_node, tree



scope_only_blocks = [
    "checked_statement",
    "fixed_statement",
    "unsafe_statement",
    "using_statement",
    "local_function_statement",
]
class_declaration_statements = [
    "constructor_declaration",
    "interface_declaration",
    "property_declaration",
    "class_declaration",
]

inner_node_type = [
    "empty_statement",
    "local_declaration_statement",
    # 'local_variable_declaration'
    # ! used in DFG graphcodebert - Swapped with local_declaration_statement
    # '_declaration',
    "labeled_statement",
    "expression_statement",
] + scope_only_blocks
non_control_statements = inner_node_type + class_declaration_statements
loop_control_statements = ["for_each_statement", "while_statement", "for_statement"]
control_statements = loop_control_statements + [
    "do_statement",
    "goto_statement",
    "throw_statement",
    "if_statement",
    "break_statement",
    "continue_statement",
    "return_statement",
    "yield_statement",
    "switch_expression",
    "switch_statement",
    "switch_expression_arm",
    "switch_section",
    "lock_statement",  # 'synchronized_statement',
    "try_statement",
    # 'try_with_resources_statement'
]
# based on handling simmilarity
statement_types = {
    "node_list_type": non_control_statements
    + control_statements
    + ["method_declaration"],
    "scope_only_blocks": scope_only_blocks,
    "outer_node_type": ["for_statement"],
    "inner_node_type": inner_node_type,
    "non_control_statement": non_control_statements,
    "control_statement": control_statements,
    "loop_control_statement": loop_control_statements,
    # 'terminal_inner': ['switch_expression_arm', 'switch_section'],
    "not_implemented": [
        # 'try_with_resources_statement'
    ],
}
# TODO: add method_return_types

def cl(child):
    if child is None:
        # logger.error
        return ""
    else:
        return child.text.decode()


def return_switch_child(node):
    """Searches for a switch descendent in the tree and returns it"""
    if node.type in ["switch_expression", "switch_statement"]:
        return node

    for child in node.children:
        t = return_switch_child(child)
        if t is not None:
            return t

    return None


def return_switch_parent(node, non_control_statement):
    """Searches for a switch expression while going up the tree and returns it"""
    while node.parent is not None:
        if node.parent.type in non_control_statement:
            return node.parent
        node = node.parent

    return None


def return_index_of_first_parent_of_type(node, parent_type):
    while node.parent is not None:
        if node.parent.type == parent_type:
            return node.parent
        node = node.parent
    return None


def get_nodes(root_node=None, node_list={}, graph_node_list=[], index={}, records={}):
    """
    Returns statement level nodes recursively from the concrete syntax tree passed to it. Uses records to maintain required supplementary information.
    noe_list maintains an intermediate representation and graph_node_list returns the final list.
    """
    #  These are within statements so handled separately - logic after assumed statements to be finest granularity
    if (
        "while" in root_node.type
        and root_node.parent is not None
        and root_node.parent.type == "do_statement"
    ):
        while_node = root_node.next_named_sibling
        label = "while(" + while_node.text.decode("UTF-8") + ")"
        type_label = "while"
        node_list[
            (while_node.start_point, while_node.end_point, while_node.type)
        ] = while_node
        graph_node_list.append(
            (
                index[(while_node.start_point, while_node.end_point, while_node.type)],
                while_node.start_point[0],
                label,
                type_label,
            )
        )
    elif root_node.type == "catch_clause":
        node_list[
            (root_node.start_point, root_node.end_point, root_node.type)
        ] = root_node
        catch_parameter = list(
            filter(lambda child: child.type == "catch_declaration", root_node.children)
        )
        label = "catch " + catch_parameter[0].text.decode("UTF-8")
        type_label = "catch"
        graph_node_list.append(
            (
                index[(root_node.start_point, root_node.end_point, root_node.type)],
                root_node.start_point[0],
                label,
                type_label,
            )
        )

    elif root_node.type == "finally_clause":
        node_list[
            (root_node.start_point, root_node.end_point, root_node.type)
        ] = root_node
        label = "finally"
        type_label = "finally"
        graph_node_list.append(
            (
                index[(root_node.start_point, root_node.end_point, root_node.type)],
                root_node.start_point[0],
                label,
                type_label,
            )
        )


    elif root_node.type in statement_types["node_list_type"]:

        if (
            root_node.type in statement_types["inner_node_type"]
            and root_node.parent is not None
            and root_node.parent.type in statement_types["outer_node_type"]
            and root_node.parent.child_by_field_name("body") != root_node
        ):
            pass
            # If it has a parent and the parent is a for loop type and it is an initialization or update statement,
            # omit it
        elif (
            root_node.type in statement_types["inner_node_type"]
            and return_switch_child(root_node) is not None
        ):
            # ? How does this logic work when there are more than 1 switch as a child
            switch_child = return_switch_child(root_node)
            child_index = index[
                (switch_child.start_point, switch_child.end_point, switch_child.type)
            ]
            current_index = index[
                (root_node.start_point, root_node.end_point, root_node.type)
            ]
            records["switch_child_map"][current_index] = child_index

        else:

            node_list[
                (root_node.start_point, root_node.end_point, root_node.type)
            ] = root_node
            # Set default label values for the node and then modify based on node type if required in the following
            # if-else ladder
            label = root_node.text.decode("UTF-8")
            type_label = "expression_statement"
            if root_node.type in [
                "method_declaration",
                "interface_declaration",
                "constructor_declaration",
                "property_declaration",
            ]:
                method_name = list(
                    filter(lambda child: child.type == "identifier", root_node.children)
                )
                parameter_list = list(
                    filter(
                        lambda child: child.type == "parameter_list", root_node.children
                    )
                )
                label = method_name[-1].text.decode("UTF-8")
                if parameter_list:
                    label = label + parameter_list[0].text.decode("UTF-8")
                type_label = root_node.type
                # ? make sure records is being updated for the newer updates if required
                records["method_list"][method_name[-1].text.decode("UTF-8")] = index[
                    root_node.start_point, root_node.end_point, root_node.type
                ]
                graph_node_list.append(
                    (
                        index[
                            (root_node.start_point, root_node.end_point, root_node.type)
                        ],
                        method_name[-1].start_point[0],
                        label,
                        type_label,
                    )
                )
            elif (
                root_node.type == "class_declaration"
                # or root_node.type == "constructor_declaration"
            ):
                modifiers = list(filter(lambda child: child.type != "declaration_list", root_node.children))
                class_name = list(filter(lambda child: child.type == "identifier", root_node.children))[0].text.decode("UTF-8")
                label = ""
                for modifier in modifiers:
                    label = label + modifier.text.decode("UTF-8") + " "
                type_label = root_node.type
                class_index = index[(root_node.start_point, root_node.end_point, root_node.type)]
                records["class_list"][class_name] = class_index

            elif root_node.type == "if_statement":
                condition = root_node.child_by_field_name("condition")
                label = f"if ( {cl(condition)} )"
                type_label = "if"
            elif root_node.type == "unsafe_statement":
                label = ""
                for child in root_node.children:
                    if not any(typ in child.type for typ in ["body", "block"]):
                        if label:
                            label = label + " "
                        label = label + child.text.decode().strip()
                type_label = "unsafe"
            elif root_node.type == "fixed_statement":
                label = ""
                for child in root_node.children:
                    if not any(typ in child.type for typ in ["body", "block"]):
                        if label:
                            label = label + " "
                        label = label + child.text.decode().strip()
                type_label = "fixed"
            elif root_node.type == "checked_statement":
                label = ""
                for child in root_node.children:
                    if not any(typ in child.type for typ in ["body", "block"]):
                        if label:
                            label = label + " "
                        label = label + child.text.decode().strip()
                type_label = "checked"
            elif root_node.type == "empty_statement":
                label = "empty"
                type_label = "empty"
            elif root_node.type == "using_statement":
                # "statement" in root_node.type and root_node.parent.type
                # ? how to handle such nodes?
                #     pass
                node_list[
                    (root_node.start_point, root_node.end_point, root_node.type)
                ] = root_node
                label = ""
                for child in root_node.children:
                    if not any(typ in child.type for typ in ["body", "block"]):
                        if label:
                            label = label + " "
                        label = label + child.text.decode().strip()
                type_label = "using"
                graph_node_list.append(
                    (
                        index[
                            (root_node.start_point, root_node.end_point, root_node.type)
                        ],
                        root_node.start_point[0],
                        label,
                        type_label,
                    )
                )

            elif root_node.type == "for_statement":
                init = root_node.child_by_field_name("initializer")
                condition = root_node.child_by_field_name("condition")
                update = root_node.child_by_field_name("update")
                label = f"for ( {cl(init)} ; {cl(condition)} ; {cl(update)} )"
                type_label = "for"

            elif root_node.type == "for_each_statement":
                l_type = root_node.child_by_field_name("type")
                l_left = root_node.child_by_field_name("left")
                l_right = root_node.child_by_field_name("right")
                label = f"foreach ( {cl(l_type)} {cl(l_left)} in {cl(l_right)} )"
                type_label = "foreach"

            elif root_node.type == "while_statement":
                label = ""
                block_hit = False
                for i, child in enumerate(root_node.children):
                    if not any(typ in child.type for typ in ["body", "block"]):
                        if not block_hit and len(root_node.children) - 1 == i:
                            continue
                        if label:
                            label = label + " "
                        label = label + child.text.decode().strip()
                type_label = "while"

            elif root_node.type == "do_statement":
                label = "do"
                type_label = "do"

            elif root_node.type == "switch_expression":
                parent_statement = return_switch_parent(
                    root_node, statement_types["non_control_statement"]
                )
                if parent_statement is not None:
                    label = parent_statement.text.decode("UTF-8").split("{")[0]
                # else:
                #     condition = list(filter(lambda child: child.type == 'parenthesized_expression', root_node.children))
                #     label = 'switch' + condition[0].text.decode('UTF-8')
                type_label = "switch_expression"

            elif root_node.type == "switch_statement":
                # parent_statement = return_switch_parent(root_node, statement_types['non_control_statement'])
                # if parent_statement is not None:
                #     label = parent_statement.text.decode('UTF-8').split('{')[0]
                # else:
                lvalue = root_node.child_by_field_name("value")
                label = f"switch ( {cl(lvalue)} )"
                type_label = "switch_statement"

            elif root_node.type == "switch_expression_arm":
                label = root_node.text.decode("UTF-8")
                type_label = "case_expression"
            elif root_node.type == "switch_section":
                type_label = "case"
                current_case_index = index[
                    (root_node.start_point, root_node.end_point, root_node.type)
                ]
                records["switch_equivalent_map"][current_case_index] = []
                parent_switch = return_index_of_first_parent_of_type(
                    root_node, "switch_statement"
                )
                parent_switch_index = index[
                    (
                        parent_switch.start_point,
                        parent_switch.end_point,
                        parent_switch.type,
                    )
                ]
                for equivalent_label in root_node.named_children:
                    if "label" in equivalent_label.type:
                        label = equivalent_label.text.decode("UTF-8").rsplit(":", 1)[0].strip()
                        label_index = index[
                            (
                                equivalent_label.start_point,
                                equivalent_label.end_point,
                                equivalent_label.type,
                            )
                        ]
                        if parent_switch_index not in records["label_switch_map"]:
                            records["label_switch_map"][parent_switch_index] = {}
                        records["label_switch_map"][parent_switch_index][
                            label
                        ] = label_index
                        records["switch_equivalent_map"][current_case_index].append(
                            label_index
                        )
                        graph_node_list.append(
                            (
                                label_index,
                                equivalent_label.start_point[0],
                                label,
                                type_label,
                            )
                        )

            elif root_node.type == "try_statement":
                # or root_node.type == 'try_with_resources_statement':
                label = "try"
                type_label = "try"

            # elif root_node.type == 'synchronized_statement':
            #     condition = list(filter(lambda child: child.type == 'parenthesized_expression', root_node.children))
            #     label = 'synchronized ' + condition[0].text.decode('UTF-8')
            #     type_label = 'synchronized'
            elif root_node.type == "labeled_statement":
                name = list(
                    filter(lambda child: child.type == "identifier", root_node.children)
                )
                label = name[0].text.decode("UTF-8").strip()
                records["label_statement_map"][label] = index[
                    (root_node.start_point, root_node.end_point, root_node.type)
                ]
                type_label = "label"

            elif root_node.type == "local_function_statement":
                label = ""
                for child in root_node.children:
                    if not any(typ in child.type for typ in ["body", "block"]):
                        if label:
                            label = label + " "
                        label = label + child.text.decode().strip()
                type_label = "local_function"
            elif root_node.type == "local_declaration_statement":
                label = ""
                for child in root_node.children:
                    if not any(typ in child.type for typ in ["body", "block"]):
                        if label:
                            label = label + " "
                        label = label + child.text.decode().strip()
                type_label = "local_declaration"
            elif root_node.type == "lock_statement":
                label = ""
                for child in root_node.children:
                    if not any(typ in child.type for typ in ["body", "block"]):
                        if label:
                            label = label + " "
                        label = label + child.text.decode().strip()
                type_label = "lock"
            # else:
            #     # TODO integrate logger.error
            #     label = ''
            #     for child in root_node.children:
            #         if not any(typ in child.type for typ in ["body", "block"]):
            #             label = label + child.text.decode().strip()
            #     type_label = root_node.type
            if root_node.type not in [
                "method_declaration",
                "constructor_declaration",
                "property_declaration",
                "switch_section",
            ]:
                graph_node_list.append(
                    (
                        index[
                            (root_node.start_point, root_node.end_point, root_node.type)
                        ],
                        root_node.start_point[0],
                        label,
                        type_label,
                    )
                )

    for child in root_node.children:
        # if child.parent.type == "local_declaration_statement":
        #     continue
        root_node, node_list, graph_node_list, records = get_nodes(
            root_node=child,
            node_list=node_list,
            graph_node_list=graph_node_list,
            index=index,
            records=records,
        )

    return root_node, node_list, graph_node_list, records



import networkx as nx


class CFGGraph:
    def __init__(self, src_language, src_code, properties, root_node, parser):
        self.src_language = src_language
        self.src_code = src_code
        self.properties = properties
        self.root_node = root_node
        self.parser = parser
        self.index = self.parser.index
        self.all_tokens = self.parser.all_tokens

    def to_networkx(self, CFG_node_list, CFG_edge_list):
        G = nx.MultiDiGraph()
        for node in CFG_node_list:
            label = str(node[1]+1) + "_ " + node[2]
            G.add_node(node[0], label=label, type_label=node[3])
        for edge in CFG_edge_list:
            additional_data = None
            if len(edge) == 4:
                additional_data = edge[3]
            if edge[2].startswith("constructor_call"):
                normal_label = "constructor_call"
            elif edge[2].startswith("method_call"):
                normal_label = "method_call"
            else:
                normal_label = edge[2]
            G.add_edge(
                edge[0],
                edge[1],
                controlflow_type=edge[2],
                edge_type="CFG_edge",
                label=normal_label,
                color="red",
            )
            if additional_data:
                for key,value in additional_data.items():
                    G.edges[edge[0], edge[1], 0][key] = value
        return G



from collections import defaultdict
import copy

import pprint

import networkx as nx

from ...utils.src_parser import traverse_tree
from ...utils.cs_nodes import statement_types
from deepdiff import DeepDiff

from loguru import logger

pp = pprint.PrettyPrinter(indent=4)
system_type = ("Console", "System", "String")
debug = False
# if any(
#     # GITHUB_ACTIONS
#     x in os.environ for x in ("PYCHARM_HOSTED",)
# ):
#     debug = True

def scope_check(parent_scope, child_scope):
    for p in parent_scope:
        if p not in child_scope:
            return False
    return True


class Identifier:
    def __init__(self, parser, node, line=None, declaration=False, full_ref=None, method_call=False):
        self.core = st(node)
        if full_ref is None:
            full_ref = node
        self.unresolved_name = st(full_ref)
        self.line = line
        self.declaration = declaration
        self.method_call = method_call
        self.satisfied = method_call
        if self.method_call and self.declaration:
            method_resolution = self.unresolved_name.split(".")
            # br.attribute2.attr.method1
            # Option 1 - method call affects main object
            # resolved_name = br
            # resolved_name = method_resolution[0]
            # resolved_name = br.attribute2.attr
            # Option 2 - method call only affects subattribute
            resolved_name = ".".join(method_resolution[:-1])
            self.name = resolved_name
        else:
            self.name = self.unresolved_name
        if not self.name:
            self.name = self.unresolved_name
        self.scope = parser.symbol_table["scope_map"][get_index(node, parser.index)]
        if line is not None:
            self.real_line_no = read_index(parser.index, line)[0][0]

    def __eq__(self, other):
        # if not isinstance(other, Identifier):
        #     return NotImplemented
        # if self.name == other.name:
        #     if other.line is None and other.line is not None:
        #         other.line = self.line
        #     elif other.line is not None and self.line is None:
        #         self.line = other.line

        # if self.method_call and self.declaration:
        #     method_resolution = self.name.split(".")
        #     # Option 1 - method call affects main object
        #     # resolved_name = method_resolution[0]
        #     # Option 2 - method call only affects subattribute
        #     resolved_name = ".".join(method_resolution[:-1])
        #     return resolved_name == other.name and self.line == other.line and sorted(self.scope) == sorted(
        #         other.scope)
        # else:
        return self.name == other.name and self.line == other.line and sorted(self.scope) == sorted(
            other.scope) and self.method_call == other.method_call

    def __hash__(self):
        # if self.method_call and self.declaration:
        #     method_resolution = self.name.split(".")
        #     # Option 1 - method call affects main object
        #     resolved_name = method_resolution[0]
        #     # Option 2 - method call only affects subattribute
        #     # resolved_name = ".".join(method_resolution[:-1])
        #     return hash((resolved_name, self.line, str(self.scope), False))
        # else:
        return hash((self.name, self.line, str(self.scope), self.method_call))

    def __str__(self):
        result = [self.name]
        if self.line:
            result += [str(self.real_line_no)]
            result += ['|'.join(map(str, self.scope))]
        else:
            result += ["?"]
        if self.method_call:
            result += ["()"]
        return f"{{{','.join(result)}}}"


def st(child):
    if child is None:
        # logger.error
        return ""
    else:
        return child.text.decode()


def get_index(node, index):
    try:
        return index[(node.start_point, node.end_point, node.type)]
    except:
        return None


def read_index(index, idx):
    return list(index.keys())[list(index.values()).index(idx)]


def parent_remapping_callback(node):
    if node is None:
        return None
    elif node.type == "do_statement":
        return None
    elif node.type == "try_statement":
        return None
        # while_node = None
        # for child in node.children:
        #     if child.type == "while":
        #         while_node = child.next_named_sibling
        #         break
        # return while_node
    else:
        return node


def return_first_parent_of_types(node, parent_types, stop_types=None):
    if stop_types is None:
        stop_types = []
    if node.type in parent_types:
        return parent_remapping_callback(node)
    while node.parent is not None:
        if node.type in stop_types:
            return None
        if node.parent.type in parent_types:
            return parent_remapping_callback(node.parent)
        node = node.parent
    return None


#
# def set_add(lst, item):
#     for entry in lst:
#         if item == entry:
#             return
#     lst.append(item)
#
#
# def set_union(first_list, second_list):
#     resulting_list = list(first_list)
#     for x in second_list:
#         set_add(resulting_list, x)
#     return resulting_list
#
#
# def set_difference(first_list, second_list):
#     resulting_list = []
#     for item in first_list:
#         match_found = False
#         for item2 in second_list:
#             if item == item2:
#                 match_found = True
#                 break
#         if not match_found:
#             resulting_list.append(item)
#     return resulting_list
#
#
# def set_intersection(first_list, second_list):
#     resulting_list = []
#     for item in first_list:
#         match_found = False
#         for item2 in second_list:
#             if item == item2:
#                 match_found = True
#                 break
#         if match_found:
#             resulting_list.append(item)
#     return resulting_list


def set_add(lst, item):
    for entry in lst:
        if item == entry:
            return
    lst.append(item)


def set_union(first_list, second_list):
    resulting_list = list(first_list)
    resulting_list.extend(x for x in second_list if x not in resulting_list)
    return resulting_list


def set_difference(first_list, second_list):
    resulting_list = [item for item in first_list if item not in second_list]
    return resulting_list


def set_intersection(first_list, second_list):
    resulting_list = [item for item in first_list if item in second_list]
    return resulting_list


def add_entry(parser, rda_table, statement_id, used=None, defined=None, declaration=False, core=None,
              method_call=False):
    if statement_id not in rda_table:
        rda_table[statement_id] = defaultdict(list)
    if not used and not defined:
        return
    if core is None:
        current_node = used or defined
        if current_node.type == "member_access_expression":
            recurse_node = recursively_get_children_of_types(current_node, ['identifier'], index=parser.index,
                                                             check_list=parser.symbol_table["scope_map"])
            if recurse_node:
                current_node = recurse_node[-1]
            if defined is not None:
                defined = current_node
            else:
                used = current_node
        if (
                current_node.parent is not None
                and current_node.parent.type == "member_access_expression"
        ):
            # LOOK AT THE SYMBOL TABLE COMMENT AS WELL #? TODO
            # object_node = current_node.parent.child_by_field_name("name")
            # object_index = get_index(object_node, parser.index)
            # current_index = get_index(current_node, parser.index)
            # if object_index == current_index:
            #     all_calls.append(current_index)

            while (
                    current_node.parent is not None
                    and current_node.parent.type == "member_access_expression"
            ):
                current_node = current_node.parent

            if (
                    current_node.parent is not None
                    and current_node.parent.type == "invocation_expression"
            ):
                if not st(current_node).startswith(system_type):
                    add_entry(parser, rda_table, statement_id, defined=used or defined, core=current_node, declaration=True,
                              method_call=True)
                    add_entry(parser, rda_table, statement_id, used=used or defined, core=current_node, declaration=True,
                              method_call=True)
                return
            if defined is not None:
                add_entry(parser, rda_table, statement_id, defined=defined, core=current_node, declaration=declaration, method_call=method_call)
            else:
                add_entry(parser, rda_table, statement_id, used=used, core=current_node, declaration=declaration, method_call=method_call)
            return
    if defined is not None:
        if get_index(defined, parser.index) not in parser.symbol_table["scope_map"]:
            return
        set_add(rda_table[statement_id]["def"],
                Identifier(parser, defined, statement_id, full_ref=core, declaration=declaration, method_call=method_call))
    else:
        if get_index(used, parser.index) not in parser.symbol_table["scope_map"]:
            return
        set_add(rda_table[statement_id]["use"],
                Identifier(parser, used, full_ref=core, declaration=declaration, method_call=method_call))


def check_field_name(req_child):
    for i, child in enumerate(req_child.parent.children):
        if req_child == child:
            field_type = child.parent.field_name_for_child(i)
            if field_type is not None and field_type == "type":
                return False
    return True


def recursively_get_children_of_types(
        node, st_types, check_list=None, index=None, result=None, stop_types=None
):
    if stop_types is None:
        stop_types = []
    if result is None:
        result = []
    if node.type in stop_types:
        return result
    if check_list and index:
        result.extend(
            list(
                filter(
                    lambda child: child.type in st_types
                                  and get_index(child, index) in check_list
                                  and check_field_name(child),
                    # and child.parent.type in ['variable_declaration', 'object_creation_expression',
                    #                           'member_access_expression'],
                    node.children,
                )
            )
        )
    else:
        result.extend(list(filter(lambda child: child.type in st_types, node.children)))
    for child in node.named_children:
        if child in stop_types:
            continue
        if child.type not in st_types:
            result = recursively_get_children_of_types(
                child, st_types, result=result, stop_types=stop_types, index=index, check_list=check_list
            )
    return result


def compute_out(old_result, connected_derivers, def_set):
    result = []
    for node in connected_derivers:
        result = set_union(result, old_result[node]["IN"])
    # for node in result:
    #     if node in def_set:
    #         continue
    return result


def compute_in(use_set, out_set, def_set, all_defs):
    result = set_union(use_set, set_difference(out_set, def_set))
    # for node in result:
    #     if node in all_defs:
    #         continue
    return result


def print_table(index, new_result, iteration=0):
    table = [f"\nRDA: iteration {iteration}\n"]
    for key in new_result:
        try:
            table.append(str(read_index(index, key)[0][0]+1)+"\n")
        except:
            table.append(str(key)+"\n")
        for key2 in new_result[key]:
            table.append(f"\t{key2}:")
            for entry in new_result[key][key2]:
                table.append(str(entry)+",")
            table.append("\n")
    logger.debug("".join(table))


def start_rda(index, rda_table, CFG_results):
    if debug:
        print_table(index, rda_table)
    cfg = CFG_results.graph
    nodes = CFG_results.graph.nodes
    old_result = {}
    iteration = 0
    for node in nodes:
        old_result[node] = {
            "IN": set(),
            "OUT": set()
        }
    # for node in nodes:
    #     connected_derivers = [t for (s, t) in cfg.out_edges(node)]
    #     old_result[node]["OUT"] = compute_out(old_result, connected_derivers, rda_table[node]["def"] if node in rda_table else [])
    new_result = copy.deepcopy(old_result)
    while True:
        iteration = iteration + 1
        for node in nodes:
            connected_feeders = [s for (s, t) in cfg.in_edges(node)]
            feeder_union = set()
            for feeder in connected_feeders:
                feeder_union = feeder_union.union(old_result[feeder]["OUT"])
            new_result[node]["IN"] = feeder_union
            def_info = rda_table[node]["def"] if node in rda_table else set()
            names_defined = [defined.name for defined in def_info]
            selective_difference = set()
            for already_defined in feeder_union:
                if already_defined.name in names_defined:
                    continue
                else:
                    selective_difference.add(already_defined)
            new_result[node]["OUT"] = selective_difference.union(def_info)
        ddiff = DeepDiff(
            old_result,
            new_result,
            ignore_order=True,
        )
        if ddiff == {}:
            # for node in nodes:
            #     all_defs = []
            #     connected_feeders = [t for (s, t) in cfg.in_edges(node)]
            #     for feeder in connected_feeders:
            #         defs = rda_table[feeder]["def"] if feeder in rda_table else []
            #         all_defs += defs
            #     for outgoing in new_result[node]["IN"]:
            #         if outgoing in all_defs:
            #             continue
            #     connected_derivers = [t for (s, t) in cfg.out_edges(node)]
            #     for deriver in connected_derivers:
            #         for incoming in new_result[deriver]["IN"]:
            #             if incoming in rda_table[node]["def"]:
            #                 continue
            if debug:
                logger.info("RDA: Completed in iteration {}", iteration)
                print_table(index, new_result, iteration)
            break
        old_result = copy.deepcopy(new_result)
    return new_result


def add_edge(final_graph, a, b, attrib=None, pre_solve=False):
    if (pre_solve and attrib) or not final_graph.has_edge(a, b):
        final_graph.add_edge(a, b)
        if attrib is not None:
            nx.set_edge_attributes(final_graph, {(a, b, 0): attrib})


def get_required_edges_from_def_to_use(index, cfg, rda_solution, rda_table, graph_nodes, properties=None):
    final_graph = copy.deepcopy(cfg)
    final_graph.remove_edges_from(list(final_graph.edges()))
    for node in graph_nodes:
        rda_info = rda_table[node] if node in rda_table else None
        if rda_info is not None:
            use_info = rda_info["use"]
            # def_info = rda_info["def"]
        else:
            use_info = set()
            # def_info = set()
        # names_used = [used.name for used in use_info]
        for used in use_info:
            for available_def in rda_solution[node]["IN"]:
                if available_def.name == used.name:
                    # if available_def in rda_table[available_def.line]["def"]:
                    if available_def.declaration:
                        if scope_check(available_def.scope, used.scope):
                            add_edge(final_graph, available_def.line, node)
                            used.satisfied = True
                    else:
                        add_edge(final_graph, available_def.line, node)
                        used.satisfied = True
                elif "." in used.name or "." in available_def.name:
                    if len(used.name.split(".")) < len(available_def.name.split(".")):
                        smaller = used
                        bigger = available_def
                    else:
                        smaller = available_def
                        bigger = used
                    if bigger.name.strip().startswith(smaller.name.strip()):
                        if not bigger.name.strip().replace(smaller.name.strip(), "").strip().startswith("."):
                            continue

                        # add_edge(final_graph, bigger.line or node, smaller.line or node)
                        add_edge(final_graph, available_def.line or node, used.line or node)
                        used.satisfied = True
            if not used.satisfied:
                if properties.get("last_use", False):
                    for i in rda_table:
                        for entry in rda_table[i]["use"]:
                            if entry.name == used.name and i != node:
                                # TODO: THIS IS AN AP - Make sure edges added here are right
                                if not nx.has_path(cfg, i, node):
                                    continue
                                add_edge(final_graph, i, node, {'color': 'green'})
        if properties.get("last_def", False):
            for available_def in rda_solution[node]["IN"] - rda_solution[node]["OUT"]:
                ignore_nodes = ['for_statement', 'while_statement', 'if_statement', 'switch_statement']
                if read_index(index, node)[-1] in ignore_nodes or read_index(index, available_def.line)[
                    -1] in ignore_nodes:
                    continue
                add_edge(final_graph, available_def.line, node, {'color': 'orange'})
        # for definition in def_info:
        #     for available_def in rda_solution[node]["IN"]:
        #         if "." in definition.name or "." in available_def.name:
        #             if len(definition.name.split(".")) < len(available_def.name.split(".")):
        #                 smaller = definition
        #                 bigger = available_def
        #             else:
        #                 smaller = available_def
        #                 bigger = definition
        #             if bigger.name.strip().startswith(smaller.name.strip()):
        #                 if not bigger.name.strip().replace(smaller.name.strip(), "").strip().startswith("."):
        #                     continue
        #                 # add_edge(final_graph, bigger.line or node, smaller.line or node)
        #                 add_edge(final_graph, smaller.line or node, bigger.line or node)

    return final_graph


def rda_cfg_map(rda_solution, CFG_results):
    graph = CFG_results.graph
    attrs = {}
    for edge in list(graph.edges):
        out_set = rda_solution[edge[0]]["OUT"]
        in_set = rda_solution[edge[1]]["IN"]
        intersection = set_intersection(out_set, in_set)
        data = graph.get_edge_data(*edge)
        if intersection:
            data['label'] = ",".join([str(intr) for intr in intersection])
        else:
            graph.remove_edge(*edge)
            # logger.warning("Unable to remap edge {}", edge)
        attrs[edge] = data
    nx.set_edge_attributes(graph, attrs)
    return graph


def dfg_csharp(properties, CFG_results):
    properties = properties
    parser = CFG_results.parser
    index = parser.index
    tree = parser.tree
    assignment = ["assignment_expression"]
    def_statement = ["variable_declarator"]
    increment_statement = ["postfix_unary_expression", "prefix_unary_expression"]
    variable_type = ['identifier', 'this_expression']
    method_calls = ["invocation_expression"]

    switch_type = ['switch_expression', 'switch_statement']
    # switch_cases = ['switch_expression_arm', 'switch_section']
    handled_cases = ["switch_section", "local_declaration_statement"]

    method_declaration = ['method_declaration']
    handled_types = assignment + def_statement + increment_statement + method_calls + method_declaration + switch_type

    # if_statement = ["if_statement", "else"]
    # for_statement = ["for_statement"]
    # enhanced_for_statement = ["for_each_statement"]
    # while_statement = ["while_statement"]

    rda_table = {}
    for root_node in traverse_tree(tree, variable_type):
        if not root_node.is_named:
            if root_node.type == "while" and root_node.parent.type == "do_statement":
                root_node = root_node.next_named_sibling
                parent_id = get_index(root_node, index)
                identifiers_used = recursively_get_children_of_types(root_node, variable_type, index=parser.index,
                                                                     check_list=parser.symbol_table["scope_map"])
                for identifier in identifiers_used:
                    add_entry(parser, rda_table, parent_id, used=identifier)
            else:
                continue
        if root_node.type == "catch_clause":
            parent_id = get_index(root_node, index)
            if len(root_node.named_children) > 1:
                identifiers_used = recursively_get_children_of_types(root_node.named_children[-2], variable_type,
                                                                     index=parser.index,
                                                                     check_list=parser.symbol_table["scope_map"])
                if identifiers_used:
                    add_entry(parser, rda_table, parent_id, defined=identifiers_used[-1])
        if root_node.type == "return_statement":
            parent_id = get_index(root_node, index)
            if len(root_node.children) < 2:
                continue
            if root_node.children[1].type == "method_invocation":
                continue
            identifiers_used = recursively_get_children_of_types(root_node, variable_type, index=parser.index,
                                                                 check_list=parser.symbol_table["scope_map"])
            for identifier in identifiers_used:
                add_entry(parser, rda_table, parent_id, used=identifier)
        if root_node.type in def_statement:
            parent_statement = return_first_parent_of_types(root_node, statement_types["node_list_type"])
            parent_id = get_index(parent_statement, index)
            if parent_id not in CFG_results.graph.nodes:
                if parent_statement and parent_statement.type in handled_cases:
                    continue
                raise NotImplementedError
            # print(parent_id)
            # print(st(parent_statement))
            if len(root_node.children) == 2:
                name = root_node.children[0]
                value = root_node.children[1]
            else:
                name = root_node.children[0]
                value = None
            add_entry(parser, rda_table, parent_id, defined=name, declaration=True)
            if value is not None:
                identifiers_used = recursively_get_children_of_types(value, variable_type, index=parser.index,
                                                                     check_list=parser.symbol_table["scope_map"])
                for identifier in identifiers_used:
                    add_entry(parser, rda_table, parent_id, used=identifier)
        elif root_node.type in assignment:
            parent_statement = return_first_parent_of_types(root_node, statement_types["node_list_type"])
            parent_id = get_index(parent_statement, index)
            # print(parent_id)
            # print(st(parent_statement))
            left_nodes = root_node.child_by_field_name("left")
            right_node = root_node.child_by_field_name("right")
            operator_text = (
                root_node.text.split(left_nodes.text, 1)[-1]
                .rsplit(right_node.text, 1)[0]
                .strip()
                .decode()
            )
            right_nodes = [right_node]
            # JRef
            if parent_id not in CFG_results.graph.nodes:
                if parent_statement and parent_statement.type in handled_cases:
                    continue
                elif parent_statement.type == "expression_statement":
                    if recursively_get_children_of_types(parent_statement, ["switch_expression"]) is not None:
                        switch_expression = recursively_get_children_of_types(parent_statement, ["switch_expression"])
                        left_nodes = switch_expression[0].parent.children[0]
                        operator_text = switch_expression[0].parent.children[1].text
                        right_nodes = []
                else:
                    raise NotImplementedError
            # operator = root_node.child()
            # operator_text = operator.text.decode('utf-8')
            if operator_text != "=":
                right_nodes.append(left_nodes)
            # TODO triage
            add_entry(parser, rda_table, parent_id, defined=left_nodes)
            for node in right_nodes:
                if node.type in variable_type:
                    add_entry(parser, rda_table, parent_id, used=node)
                else:
                    identifiers_used = recursively_get_children_of_types(node, variable_type, index=parser.index,
                                                                         check_list=parser.symbol_table["scope_map"])
                    for identifier in identifiers_used:
                        add_entry(parser, rda_table, parent_id, used=identifier)
        elif root_node.type in increment_statement:
            parent_statement = return_first_parent_of_types(root_node, statement_types["node_list_type"])
            parent_id = get_index(parent_statement, index)
            if parent_id not in CFG_results.graph.nodes:
                if parent_statement and parent_statement.type in handled_cases:
                    continue
                raise NotImplementedError
            # print(parent_id)
            # print(st(parent_statement))
            if root_node.type in variable_type:
                add_entry(parser, rda_table, parent_id, used=root_node)
            else:
                identifiers_used = recursively_get_children_of_types(root_node, variable_type, index=parser.index,
                                                                     check_list=parser.symbol_table["scope_map"])
                for identifier in identifiers_used:
                    add_entry(parser, rda_table, parent_id, used=identifier)
                    add_entry(parser, rda_table, parent_id, defined=identifier)
        elif root_node.type in method_calls:
            parent_statement = return_first_parent_of_types(root_node, statement_types["node_list_type"])
            parent_id = get_index(parent_statement, index)
            if parent_id not in CFG_results.graph.nodes:
                if parent_statement and parent_statement.type in handled_cases:
                    continue
                raise NotImplementedError
            add_entry(parser, rda_table, parent_id, used=root_node.children[0], method_call=True)
            for node in root_node.children[1:]:
                if node.type in variable_type:
                    if return_first_parent_of_types(node, 'invocation_expression') is None:
                        add_entry(parser, rda_table, parent_id, used=node)
                else:
                    identifiers_used = recursively_get_children_of_types(node, variable_type, index=parser.index,
                                                                         check_list=parser.symbol_table["scope_map"])
                    for identifier in identifiers_used:
                        add_entry(parser, rda_table, parent_id, used=identifier)
        elif root_node.type in method_declaration:
            parent_id = get_index(root_node, index)
            for child in root_node.named_children:
                if child.type == "parameter_list":
                    for parameter in child.named_children:
                        if parameter.type == "parameter":
                            identifier = parameter.child_by_field_name("name")
                            if identifier.parent.type == "identifier" and not identifier.text:
                                identifier = identifier.parent
                            add_entry(parser, rda_table, parent_id, defined=identifier)
                    break
        elif root_node.type in switch_type:
            parent_statement = root_node
            parent_id = get_index(root_node, index)
            if parent_id not in CFG_results.graph.nodes:
                if parent_statement and parent_statement.type in handled_cases:
                    continue
                raise NotImplementedError
            if parent_statement.type == 'switch_expression':
                add_entry(parser, rda_table, parent_id, used=root_node.children[0])
            elif parent_statement.type == 'switch_statement':
                case_node = parent_statement.child_by_field_name("value")
                add_entry(parser, rda_table, parent_id, used=case_node)
        elif root_node.type == "for_each_statement":
            parent_statement = root_node
            parent_id = get_index(root_node, index)
            defined_node = parent_statement.child_by_field_name("left")
            used_node = parent_statement.child_by_field_name("right")
            add_entry(parser, rda_table, parent_id, defined=defined_node, declaration=True)
            add_entry(parser, rda_table, parent_id, used=used_node)
        else:
            # THIS ELSE IS VERY INEFFICIENT
            # Essentially considers all variable references as use
            # if return_first_parent_of_types(root_node, handled_types) is None:
            parent_statement = return_first_parent_of_types(
                root_node,
                # statement_types["node_list_type"],
                # [x for x in statement_types["node_list_type"] if x not in block_types],
                statement_types["non_control_statement"] + statement_types["control_statement"],
                stop_types=['block'] + handled_types
            )
            parent_id = get_index(parent_statement, index)
            if parent_id is None:
                continue
            if parent_id not in CFG_results.graph.nodes:
                if parent_statement and parent_statement.type in handled_cases:
                    continue
                raise NotImplementedError
            # print(root_node.type)
            # print(parent_statement.type)

            # print("...")

            # print(parent_id)
            # print(st(parent_statement))

            # if parent_statement and parent_statement.type in block_types:
            #     continue

            # if root_node.type in variable_type:
            #     add_entry(parser, rda_table, parent_id, used=root_node)
            # else:
            identifiers_used = recursively_get_children_of_types(root_node, variable_type, stop_types=handled_types,
                                                                 index=parser.index,
                                                                 check_list=parser.symbol_table["scope_map"])
            for identifier in identifiers_used:
                add_entry(parser, rda_table, parent_id, used=identifier)
    rda_solution = start_rda(index, rda_table, CFG_results)
    # pp.pprint(rda_solution)
    final_graph = get_required_edges_from_def_to_use(CFG_results.parser.index, CFG_results.graph, rda_solution, rda_table,
                                                     CFG_results.graph.nodes, properties=properties)
    debug_graph = rda_cfg_map(rda_solution, CFG_results)
    return final_graph, debug_graph, rda_table, rda_solution



import os.path
import time

import networkx as nx
from loguru import logger

from ...codeviews.CFG.CFG_driver import CFGDriver
from .SDFG_csharp import dfg_csharp
from .SDFG_java import dfg_java
from ...utils import postprocessor, DFG_utils

debug = False


# if any(
#         # GITHUB_ACTIONS
#         x in os.environ for x in ("PYCHARM_HOSTED",)
# ):
#     debug = True

class DfgRda:
    def __init__(
            self,
            src_language="java",
            src_code="",
            output_file=None,
            graph_format="dot",
            properties=None,
    ):
        if not properties:
            properties = {
                "CFG": {},
                "DFG": {
                    "last_use": True,
                    "last_def": True,
                    "alex_algo": True,
                },
            }
        self.src_language = src_language
        self.src_code = src_code
        self.properties = properties
        self.graph = nx.MultiDiGraph()
        # time to create the CFG
        start = time.time()
        self.CFG_Results = CFGDriver(
            self.src_language, self.src_code, "", self.properties["CFG"]
        )
        end = time.time()
        self.CFG = self.CFG_Results.graph
        self.rda_table = None
        self.rda_result = None
        # time to create the DFG
        start_dfg = time.time()
        self.graph, self.debug_graph, self.rda_table, self.rda_result = self.rda(
            self.properties["DFG"]
        )
        end_dfg = time.time()
        if debug:
            logger.warning("CFG time: " + str(end - start) + " DFG time: " + str(end_dfg - start_dfg))
        if output_file:
            if graph_format == "all" or graph_format == "json":
                self.json = postprocessor.write_networkx_to_json(
                    self.graph, output_file
                )
            if graph_format == "all" or graph_format == "dot":
                postprocessor.write_to_dot(
                    self.graph, output_file.rsplit(".", 1)[0] + ".dot", output_png=True
                )
            if graph_format == "all" or graph_format == "dot":
                postprocessor.write_to_dot(
                    self.debug_graph, output_file.rsplit(".", 1)[0] + "_debug.dot", output_png=True
                )
            self.json = postprocessor.write_networkx_to_json(self.graph, output_file)

    def get_graph(self):
        return self.graph

    def index_to_code(self):
        tokens_index = DFG_utils.tree_to_token_index(self.CFG_Results.root_node)
        code = self.src_code.split("\n")
        # code_tokens is a list of all tokens in code - all leaves of the CST
        code_tokens = [DFG_utils.index_to_code_token(x, code) for x in tokens_index]
        index_to_code = {}

        for (ind, code) in zip(tokens_index, code_tokens):
            if ind in self.CFG_Results.parser.index:
                idx = self.CFG_Results.parser.index[ind]
            else:
                idx = -1
            index_to_code[ind] = (idx, code)

        return index_to_code

    def rda(self, properties):
        lang_map = {
            "java": dfg_java,
            "cs": dfg_csharp,
        }
        driver = lang_map[self.src_language]
        return driver(properties, self.CFG_Results)


if __name__ == '__main__':
    # file = "tests/data/RANDOM/translation.cs"
    # # sample_id = random.randint(0, 11800 - 1)
    # sample_id = 11132
    # with open(file, "r") as f:
    #     lines = f.read().splitlines()
    # sample_file = lines[sample_id]

    # "cs",
    for extension in ("java", "cs"):
        file = f"data/test_manual.{extension}"
        if not os.path.isfile(file):
            continue
        with open(file, "r") as f:
            sample_file = f.read()

        output_file = f"data/{extension}_test.json"
        # src_code = src_parser.pre_process_src(extension, sample_file, wrap_class=False)
        src_code = sample_file
        # with open("data/test." + extension, "w") as f:
        #     f.write(src_code)
        result = DfgRda(
            src_language=extension,
            src_code=src_code,
            output_file=None,
        )
        graph = result.graph
        postprocessor.write_to_dot(
            graph,
            output_file.rsplit(".", 1)[0] + ".dot",
            output_png=True,
        )
        debug_graph = result.debug_graph
        postprocessor.write_to_dot(
            debug_graph,
            output_file.rsplit(".", 1)[0] + "_debug.dot",
            output_png=True,
        )



import os

import networkx as nx

from ..AST.AST_driver import ASTDriver
from ..CFG.CFG_driver import CFGDriver
from ..DFG.DFG_driver import DFGDriver
from ...utils import postprocessor


class CombinedDriver:
    def __init__(
        self,
        src_language="java",
        src_code="",
        output_file=None,
        graph_format="dot",
        codeviews={},
    ):
        self.src_language = src_language
        self.src_code = src_code
        self.codeviews = codeviews
        self.graph = nx.MultiDiGraph()
        self.results = {}

        if self.codeviews["DFG"]["exists"] == True:
            self.results["DFG"] = DFGDriver(
                self.src_language, self.src_code, "", self.codeviews
            )
            self.DFG = self.results["DFG"].graph

        if self.codeviews["AST"]["exists"] == True:
            self.results["AST"] = ASTDriver(
                self.src_language, self.src_code, "", self.codeviews["AST"]
            )
            self.AST = self.results["AST"].graph

        if self.codeviews["CFG"]["exists"] == True:
            self.results["CFG"] = CFGDriver(
                self.src_language, self.src_code, "", self.codeviews["CFG"]
            )
            self.CFG = self.results["CFG"].graph

        self.combine()
        if output_file:
            if graph_format == "all" or graph_format == "json":
                self.json = postprocessor.write_networkx_to_json(
                    self.graph, output_file
                )
            if graph_format == "all" or graph_format == "dot":
                postprocessor.write_to_dot(
                    self.graph, output_file.split(".")[0] + ".dot", output_png=True
                )

    def get_graph(self):
        return self.graph

    def check_validity(self):
        """Write logic for valid combinations here"""
        return True

    def AST_simple(self):
        self.graph = self.AST
        grandparent_folder = os.path.abspath(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        )
        # postprocessor.write_to_dot(self.graph, os.path.join(grandparent_folder,"output_graphs/AST_output.dot"))
        # postprocessor.write_networkx_to_json(self.graph, os.path.join(grandparent_folder,"./output_json/AST_output.json"))

    def DFG_simple(self):
        self.graph = self.DFG
        grandparent_folder = os.path.abspath(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        )
        # postprocessor.write_to_dot(self.graph, os.path.join(grandparent_folder,"output_graphs/DFG_output.dot"))
        # postprocessor.write_networkx_to_json(self.graph, os.path.join(grandparent_folder,"./output_json/DFG_output.json"))

    def CFG_simple(self):
        self.graph = self.CFG
        grandparent_folder = os.path.abspath(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        )
        # postprocessor.write_to_dot(self.graph, os.path.join(grandparent_folder,"output_graphs/CFG_output.dot"))
        # postprocessor.write_networkx_to_json(self.graph, os.path.join(grandparent_folder,"output_json/CFG_output.json"))

    def DFG_collapsed(self):
        self.graph = self.DFG
        grandparent_folder = os.path.abspath(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        )
        # postprocessor.write_to_dot(self.graph, os.path.join(grandparent_folder,"output_graphs/DFG_collapsed_output.dot"))
        # postprocessor.write_networkx_to_json(self.graph, os.path.join(grandparent_folder,"output_json/DFG__collapsed_output.json"))

    def AST_collapsed(self):
        self.graph = self.AST
        grandparent_folder = os.path.abspath(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        )
        # postprocessor.write_to_dot(self.graph, os.path.join(grandparent_folder,"output_graphs/AST_collapsed_output.dot"))
        # postprocessor.write_networkx_to_json(self.graph, os.path.join(grandparent_folder,"output_json/AST_collapsed_output.json"))

    def combine_AST_DFG_simple(self):
        self.graph.add_nodes_from(self.AST.nodes(data=True))
        self.graph.add_nodes_from(self.DFG.nodes(data=True))
        self.graph.add_edges_from(self.AST.edges(data=True))
        self.graph.add_edges_from(self.DFG.edges(data=True))

        grandparent_folder = os.path.abspath(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        )
        # postprocessor.write_to_dot(self.graph, os.path.join(grandparent_folder,"output_graphs/AST_DFG_simple_output.dot"))
        # postprocessor.write_networkx_to_json(self.graph, os.path.join(grandparent_folder,"output_json/AST_DFG_simple_output.json"))

    def combine_CFG_DFG_simple(self):
        self.graph.add_nodes_from(self.CFG.nodes(data=True))
        self.graph.add_nodes_from(self.DFG.nodes(data=True))
        self.graph.add_edges_from(self.CFG.edges(data=True))
        self.graph.add_edges_from(self.DFG.edges(data=True))

        grandparent_folder = os.path.abspath(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        )
        # postprocessor.write_to_dot(self.graph, os.path.join(grandparent_folder,"output_graphs/CFG_DFG_simple_output.dot"))
        # postprocessor.write_networkx_to_json(self.graph, os.path.join(grandparent_folder,"output_json/CFG_DFG_simple_output.json"))

    def combine_AST_CFG_simple(self):
        self.graph.add_nodes_from(self.AST.nodes(data=True))
        self.graph.add_nodes_from(self.CFG.nodes(data=True))
        self.graph.add_edges_from(self.AST.edges(data=True))
        self.graph.add_edges_from(self.CFG.edges(data=True))
        grandparent_folder = os.path.abspath(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        )
        # postprocessor.write_to_dot(self.graph, os.path.join(grandparent_folder,"output_graphs/AST_CFG_simple_output.dot"))
        # postprocessor.write_networkx_to_json(self.graph, os.path.join(grandparent_folder,"output_json/AST_CFG_simple_output.json"))

    def combine_AST_CFG_DFG_simple(self):
        self.graph.add_nodes_from(self.AST.nodes(data=True))
        self.graph.add_nodes_from(self.CFG.nodes(data=True))
        self.graph.add_nodes_from(self.DFG.nodes(data=True))
        self.graph.add_edges_from(self.AST.edges(data=True))
        self.graph.add_edges_from(self.CFG.edges(data=True))
        self.graph.add_edges_from(self.DFG.edges(data=True))
        grandparent_folder = os.path.abspath(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        )
        # postprocessor.write_to_dot(self.graph, os.path.join(grandparent_folder,"output_graphs/AST_CFG_DFG_simple_output.dot"))
        # postprocessor.write_networkx_to_json(self.graph, os.path.join(grandparent_folder,"output_json/AST_CFG_DFG_simple_output.json"))

    def combine_AST_DFG_collapsed(self):
        self.graph.add_nodes_from(self.AST.nodes(data=True))
        self.graph.add_nodes_from(self.DFG.nodes(data=True))
        self.graph.add_edges_from(self.AST.edges(data=True))
        self.graph.add_edges_from(self.DFG.edges(data=True))
        grandparent_folder = os.path.abspath(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        )
        # postprocessor.write_to_dot(self.graph, os.path.join(grandparent_folder,"output_graphs/AST_DFG_collapsed_output.dot"))
        # postprocessor.write_networkx_to_json(self.graph, os.path.join(grandparent_folder,"output_json/AST_DFG_collapsed_output.json"))

    def combine(self):
        """Combine all combinations into a single graph"""

        if (
            self.codeviews["AST"]["exists"] == True
            and self.codeviews["CFG"]["exists"] == True
            and self.codeviews["DFG"]["exists"] == True
        ):
            # if self.codeviews["DFG"]["collapsed"] == False and self.codeviews["AST"]["collapsed"] == False:
            self.combine_AST_CFG_DFG_simple()

        elif (
            self.codeviews["AST"]["exists"] == True
            and self.codeviews["DFG"]["exists"] == True
        ):
            if (
                self.codeviews["DFG"]["collapsed"] == False
                and self.codeviews["AST"]["collapsed"] == False
            ):
                self.combine_AST_DFG_simple()

            elif (
                self.codeviews["DFG"]["collapsed"] == True
                and self.codeviews["AST"]["collapsed"] == True
            ):
                self.combine_AST_DFG_collapsed()

        elif (
            self.codeviews["AST"]["exists"] == True
            and self.codeviews["CFG"]["exists"] == True
        ):
            # if self.codeviews["DFG"]["collapsed"] == False and self.codeviews["AST"]["collapsed"] == False:
            self.combine_AST_CFG_simple()

        elif (
            self.codeviews["CFG"]["exists"] == True
            and self.codeviews["DFG"]["exists"] == True
        ):
            # if self.codeviews["DFG"]["collapsed"] == False and self.codeviews["AST"]["collapsed"] == False:
            self.combine_CFG_DFG_simple()

        elif self.codeviews["AST"]["exists"] == True:
            if self.codeviews["AST"]["collapsed"] == True:
                self.AST_collapsed()
            else:
                self.AST_simple()

        elif self.codeviews["DFG"]["exists"] == True:
            if self.codeviews["DFG"]["collapsed"] == True:
                self.DFG_collapsed()
            else:
                self.DFG_simple()
        elif self.codeviews["CFG"]["exists"] == True:

            self.CFG_simple()



# This code is from https://github.com/microsoft/CodeBERT/blob/master/GraphCodeBERT/clonedetection/parser/utils.py

import re
import tokenize
from io import StringIO


def remove_empty_lines(source):
    temp = []
    for x in source.split("\n"):
        # if x.strip() != "":
        temp.append(x)
    return "\n".join(temp)


def remove_comments(lang, source):
    if lang in ["python"]:
        """
        Returns 'source' minus comments and docstrings.
        """
        io_obj = StringIO(source)
        out = ""
        prev_toktype = tokenize.INDENT
        last_lineno = -1
        last_col = 0
        for tok in tokenize.generate_tokens(io_obj.readline):
            token_type = tok[0]
            token_string = tok[1]
            start_line, start_col = tok[2]
            end_line, end_col = tok[3]
            ltext = tok[4]
            if start_line > last_lineno:
                last_col = 0
            if start_col > last_col:
                out += " " * (start_col - last_col)
            # Remove comments:
            if token_type == tokenize.COMMENT:
                pass
            # This series of conditionals removes docstrings:
            elif token_type == tokenize.STRING:
                if prev_toktype != tokenize.INDENT:
                    # This is likely a docstring; double-check we're not inside an operator:
                    if prev_toktype != tokenize.NEWLINE:
                        if start_col > 0:
                            out += token_string
            else:
                out += token_string
            prev_toktype = token_type
            last_col = end_col
            last_lineno = end_line
        temp = []
        for x in out.split("\n"):
            if x.strip() != "":
                temp.append(x)
        return "\n".join(temp)
    elif lang in ["ruby"]:
        return source
    else:

        def replacer(match):
            s = match.group(0)
            if s.startswith("/"):
                return " "  # note: a space and not an empty string
            else:
                return s

        pattern = re.compile(
            r'//.*?$|/\*.*?\*/|\'(?:\\.|[^\\\'])*\'|"(?:\\.|[^\\"])*"',
            re.DOTALL | re.MULTILINE,
        )
        temp = []
        for x in re.sub(pattern, replacer, source).split("\n"):
            # if x.strip() != "":
            temp.append(x)
            # else:
        return "\n".join(temp)



import networkx as nx


class ASTGraph:
    def __init__(self, src_language, src_code, properties, root_node, parser):
        self.src_language = src_language
        self.src_code = src_code
        self.properties = properties
        self.root_node = root_node
        self.parser = parser
        self.index = self.parser.index
        self.all_tokens = self.parser.all_tokens
        self.label = self.parser.label
        self.method_map = self.parser.method_map
        self.method_calls = self.parser.method_calls
        self.start_line = self.parser.start_line
        self.graph = self.to_networkx()

    def get_AST_nodes(self, root_node, AST, AST_index):

        if root_node.is_named:
            current_node_id = AST_index[
                (root_node.start_point, root_node.end_point, root_node.type)
            ]
            if current_node_id in self.all_tokens:

                label = self.label[current_node_id]
            else:
                label = root_node.type
            AST.add_node(
                current_node_id,
                node_type=root_node.type,
                label=label,
                shape="box",
                style="rounded, filled",
                fillcolor="#BFE6D3",
                color="white",
            )
            for child in root_node.children:
                if child.is_named:
                    child_id = self.get_AST_nodes(child, AST, AST_index)
                    if child_id != None:
                        AST.add_edge(current_node_id, child_id)
            return current_node_id

    def merge_nodes(self, G, nodes, new_node):
        """
        Merges the selected `nodes` of the graph G into one `new_node`,
        meaning that all the edges that pointed to or from one of these
        `nodes` will point to or from the `new_node`.
        """

        nodes.remove(new_node)

        incoming_edges = G.in_edges(nodes, data=True)
        outgoing_edges = G.out_edges(nodes, data=True)
        for i in incoming_edges:
            G.add_edge(i[0], new_node, **i[2])
        for i in outgoing_edges:
            G.add_edge(new_node, i[1], **i[2])

        for n in nodes:  # remove the merged nodes
            G.remove_node(n)

    def collapse(self, G):
        name_to_index_map = {}
        node_attributes = {}

        for variable in self.all_tokens:
            if variable not in self.method_map:
                name_to_index_map[self.label[variable]] = set()

        for node, properties in G.nodes(data=True):
            name = properties["label"]
            if name in name_to_index_map.keys():
                name_to_index_map[name].add(node)

        for name, indexes in name_to_index_map.items():
            node_list = list(indexes)
            chosen_node = min(node_list)
            node_attributes[chosen_node] = G.nodes(data=True)[chosen_node]
            self.merge_nodes(G, node_list, chosen_node)
        nx.set_node_attributes(G, node_attributes)

    def minimize(self, root_node, blacklisted_nodes):
        if root_node.type in self.properties.get("blacklisted", []):
            blacklisted_nodes.append(
                self.index[(root_node.start_point, root_node.end_point, root_node.type)]
            )

        for child in root_node.children:
            self.minimize(child, blacklisted_nodes)

        return blacklisted_nodes

    def remove_blacklisted_nodes(self, G):
        blacklisted_nodes = self.minimize(self.root_node, [])

        for node in blacklisted_nodes:
            for predecessor in G.predecessors(node):
                for successor in G.successors(node):
                    G.add_edge(predecessor, successor)

            G.remove_node(node)

    def to_networkx(self):
        G = nx.MultiDiGraph()
        self.get_AST_nodes(self.root_node, G, self.index)
        if self.properties.get("collapsed", False) == True:
            self.collapse(G)
        if self.properties.get("minimized", False) == True:
            self.remove_blacklisted_nodes(G)

        nx.set_edge_attributes(G, "AST_edge", "edge_type")
        nx.set_edge_attributes(G, "indigo", "color")
        nx.set_edge_attributes(G, "vee", "shape")
        return G



import os

import networkx as nx
from ...tree_parser.parser_driver import ParserDriver


class CSTGraph:
    def __init__(self, src_language, src_code, root_node):
        self.src_language = src_language
        self.src_code = src_code
        self.root_node = root_node
        self.index = {}
        self.create_CST_id(self.root_node, self.index, [0])
        self.graph = self.to_networkx()
        grandparent_folder = os.path.abspath(
            os.path.dirname(os.path.dirname(os.path.dirname(__file__)))
        )
        # postprocessor.write_to_dot(self.graph, os.path.join(grandparent_folder,"output_graphs/CST_output.dot"))

    def postorder_traversal(self, node):
        if node.children:
            for child in node.children:
                self.postorder_traversal(child)
        print(node.type)

    def create_CST_id(self, root_node, CST_index, CST_id):
        current_node_id = CST_id[0]
        CST_id[0] += 1
        CST_index[
            (root_node.start_point, root_node.end_point, root_node.type)
        ] = current_node_id
        for child in root_node.children:

            child_id = self.create_CST_id(child, CST_index, CST_id)
        return current_node_id  # redundant

    def get_CST_nodes(self, root_node, CST, CST_index):
        current_node_id = CST_index[
            (root_node.start_point, root_node.end_point, root_node.type)
        ]
        label = str(root_node.start_point[0]) + "_" + root_node.type
        CST.add_node(
            current_node_id,
            label=label,
            shape="box",
            style="rounded, filled",
            fillcolor="#BFE6D3",
            color="white",
        )
        for child in root_node.children:

            child_id = self.get_CST_nodes(child, CST, CST_index)
            if child_id != None:
                CST.add_edge(current_node_id, child_id)
        return current_node_id

    def to_networkx(self):
        G = nx.MultiDiGraph()
        self.get_CST_nodes(self.root_node, G, self.index)
        nx.set_edge_attributes(G, "CST_edge", "edge_type")
        nx.set_edge_attributes(G, "blue", "color")
        nx.set_edge_attributes(G, "vee", "shape")
        return G


class CSTDriver:
    def __init__(self, src_language="java", src_code=""):
        self.src_language = src_language
        self.src_code = src_code
        self.parser = ParserDriver(src_language, src_code).parser
        self.root_node, self.tree = self.parser.parse()
        self.graph = CSTGraph(self.src_language, self.src_code, self.root_node)



import json
import os
import re
from subprocess import check_call

import networkx as nx
from networkx.readwrite import json_graph


def networkx_to_json(graph):
    """Convert a networkx graph to a json object"""
    graph_json = json_graph.node_link_data(graph)
    return graph_json


def write_networkx_to_json(graph, filename):
    """Convert a networkx graph to a json object"""
    graph_json = json_graph.node_link_data(graph)
    if not os.getenv("GITHUB_ACTIONS"):
        with open(filename, "w") as f:
            json.dump(graph_json, f)
    return graph_json


def to_dot(graph):
    return nx.nx_pydot.to_pydot(graph)


def write_to_dot(graph, filename, output_png=False):
    if not os.getenv("GITHUB_ACTIONS"):
        for node in graph.nodes:
            if 'label' in graph.nodes[node]:
                graph.nodes[node]['label'] = re.escape(graph.nodes[node]['label'])
        nx.nx_pydot.write_dot(graph, filename)
        if output_png:
            check_call(
                ["dot", "-Tpng", filename, "-o", filename.rsplit(".", 1)[0] + ".png"]
            )



from comex.tree_parser.parser_driver import ParserDriver
from comex.utils.cs_nodes import statement_types as cs_statement_types
from comex.utils.java_nodes import statement_types as java_statement_types


def traverse_tree(tree, finest_granularity=None):
    if finest_granularity is None:
        finest_granularity = []
    cursor = tree.walk()

    reached_root = False
    while not reached_root:
        yield cursor.node

        if cursor.goto_first_child() and cursor.node.type not in finest_granularity:
            continue

        if cursor.goto_next_sibling():
            continue

        retracing = True
        while retracing:
            if not cursor.goto_parent():
                retracing = False
                reached_root = True

            if cursor.goto_next_sibling():
                retracing = False


def pre_process_src(extension, line, wrap_class=False, ignore_error=False):
    if wrap_class:
        line = "public class test {" + line + "}"
    parsed = ParserDriver(extension, line)
    # new_line_pos = set()
    fixed = ""
    if extension == "java":
        statement_types = java_statement_types
    else:
        statement_types = cs_statement_types
    if not ignore_error:
        if parsed.parser.root_node.has_error:
            return None
        # for tag in parsed.parser.index:
        #     if tag[2] == "ERROR":
        #         return None
        # if tag[2] in statement_types["node_list_type"]:
        #     new_line_pos.add(tag[0][-1]-1)
    # for pos, char in enumerate(line):
    #     fixed = fixed + char
    #     if pos in new_line_pos:
    #         fixed = fixed + "\n"
    for node in traverse_tree(parsed.tree):
        # ? keeping on node.children check causes literals to be dropped case 7129
        if node.type in statement_types["node_list_type"]:
            fixed = fixed + "\n"
        if (
                not node.children and "literal" not in node.parent.type
        ) or "literal" in node.type:
            fixed = fixed + node.text.decode() + " "

    return fixed
